// This may look like C code, but it is really -*- C++ -*-
//
// Regular Transient analysis. 
// 
// by Lei Zhu and Carlos Christoffersen
//
// 

#include "global.h"
#include <carrot/Circuit.h>
#include "EnvelopeHB.h"
#include <carrot/NewtonSolve.h>
#include "DVectorHolder.h"

#include <stdlib.h>
#include <cmath>
#include <cstdio>
#include <string>

// Debug
//#define DEBUG 1

static ItemInfo EnvelopeHB_info = {
  "ehb",
  "Envelope/Normal Harmonic Balance Analysis",
  "Lei Zhu, Carlos Christoffersen",
  DEFAULT_ADDRESS
};

enum {TSTOP, TSTEP, FREQ, MAX_DPARM_INDEX}; 

static ParamHandler<double>::ParmInfo EnvelopeHB_dparams[] = {
  {"tstop", "Stop time (s)", true, 0.},
  {"tstep", "Time step value (s)", true, 0.},
  {"freq", "The fundamental frequency", true, 1000.}
};

// ***********************************************************************
// Integer parameters
// ***********************************************************************
enum {NH, SENS, NORM, OUTSTEPS, OVERSAMPLING, MAX_IPARM_INDEX};

static ParamHandler<int>::ParmInfo EnvelopeHB_iparams[] = {
  {"nh", "Number of Harmonics", false, 8},  
  {"sens", "Perform Sensitivity analysis (0: no)", false, 0},
  {"norm", "Calculate normalised sensitivities if possible (0: no)", false, 1},
  {"outsteps", "Number of steps to show simulation progress", false, 10},
  {"oversampling", "Number of extra samples in FFT (even, min. 2)", false, 8}
};

// ***********************************************************************
// String parameters
// ***********************************************************************
enum {NODENAME, MODELNAME, MAX_SPARM_INDEX};

static ParamHandler<std::string>::ParmInfo EnvelopeHB_sparams[] = {
  {"nodename", "Name of the node to calculate sensitivities", false, "out1"},
  {"modelname", "Sensitivities are calculated with respect to this model", 
   false, "all"}
};

extern "C" {

  //Export the name of the module we are using
  const char * AnalysisName(){
    return EnvelopeHB_info.type.c_str();
  }
  ParmSet * newParmSet(const char * model){
    return new ParmSet(model, &EnvelopeHB_info, 
		       EnvelopeHB_dparams, MAX_DPARM_INDEX, 
		       EnvelopeHB_iparams, MAX_IPARM_INDEX,
		       EnvelopeHB_sparams, MAX_SPARM_INDEX);
  }
  void delParmSet(ParmSet * p){
    delete p;
    p = 0;
  }
  Analysis * newObject(ParmSet * p){
    return new EnvelopeHB(p);
  }
  void delObject(Analysis * e){
    delete e;
    e = 0;
  }
};

void EnvelopeHB::run(Circuit* cir, OutputManager* om)
{
  // Check parameters
  getParmSet()->checkParms();

  // Set variables in OutputManager
  om->setVarUnit("s"); // Seconds
  om->setAnalysis("EHB_td");

  double *dparm = getParmSet()->getDHandler()->getParmVec();
  int *iparm = getParmSet()->getIHandler()->getParmVec();
//   std::string *sparm = getParmSet()->getSHandler()->getParmVec();

  printf("****************************************************************\n");
  printf("*                       EnvelopeHB Analysis                    *\n");
  printf("****************************************************************\n");

  // Set row/column numbers in terminals and get number of EnvelopeHB-used
  // terminals. 
  int nterms=cir->setMatrixRCs();
  // Get rows of desired output variables
  std::vector<int> iv;
  om->getRowNumbers(cir, iv);

  // ***********************  Create matrix and vectors
  // Ground node not included in nrcs
  nrcs=nterms-1; 

  // pointer for each current source type vector
  linvccs = cir->getLinVCCSvec();
  LinfVCCSVector linfvccs = cir->getLinfVCCSvec();
  linvcqs = cir->getLinVCQSvec();
  indcs = cir->getIndepCSvec();
  TDCSVector tdcs = cir->getTDCSvec();
  genvccs = cir->getGenVCCSvec();
  genvcqs = cir->getGenVCQSvec();

  num_HB = iparm[NH]; 
  // Angular frequencies vector
  double omega0=twopi*dparm[FREQ];
  Omega = new double [num_HB];
  Omega[0] = 0.;
  for (int i(1); i<num_HB; i++)
    Omega[i] = Omega[i-1] + omega0;

  // vector of nodal voltages in frequency domain
  double_complex *uvec_ground = new double_complex [num_HB*nterms];
  double_complex *svec_ground = new double_complex [num_HB*nterms];
  // Clear vectors (this sets initial conditions to zero for now)
  for (int j(0); j < num_HB*nterms; j++){
    uvec_ground[j] = svec_ground[j] = 0.;
  }
  uvec = uvec_ground+num_HB;
  svec = svec_ground+num_HB;
  // Nonlinear charge sources:
  nCharges = genvcqs.size();
  q = new double_complex[nCharges*num_HB];
  // Set initial conditions in VCQS
  for (int j(0); j<nCharges*num_HB; j++)
    q[j] = 0.;

  int fsize = 2*num_HB - 1;
  UMFPackDI nam(fsize*nrcs, 24*fsize*nrcs);
  double* uvec_d = new double [fsize*nrcs];
  double* svec_d = new double [fsize*nrcs];

  printf("\nNumber of nodes (excluding reference): %d\n", nrcs);
  printf("\nSystem dimension: %d\n", fsize*nrcs);
  printf("\n Time \t\t Time Steps \t Average Iter. \t Max. Residual\n");
  printf("---------------------------------------------------------------\n");

  //fft and ifft (fftw3)
  unsigned flags(FFTW_ESTIMATE);
  int ovspl = iparm[OVERSAMPLING];
  if (ovspl < 2) {
    printf("\nWarning: oversampling set to 2\n");
    ovspl = 2;
  }
  // Make sure ovspl is even
  if (ovspl % 2)
    ovspl++;
  N = 2*(num_HB-1) + ovspl;
  padsize = N/2+1;
  put1 = new double [N];                     // time domain
  put2 = new double_complex [padsize];        // frequency domain
  p1 = fftw_plan_dft_r2c_1d(N, put1, reinterpret_cast<fftw_complex*>(put2), 
			    flags);   // fft
  p2 = fftw_plan_dft_c2r_1d(N, reinterpret_cast<fftw_complex*>(put2), 
			    put1, flags);   // ifft
  
  // Create second time-axis vector
  double* time2 = new double [N];
  double h2 = 1./dparm[FREQ]/N; // Grid spacing in t2 direction
  time2[0] = 0.;
  for (int i(1); i<N; i++)
    time2[i] = time2[i-1] + h2;

  // Nonlinear current sources
  const int &max_cv = cir -> getMaxCV();  // maximum number of control voltages
  double *uvec_td_ground = new double [N*nterms];  
  for (int i(0); i<N; i++)
    uvec_td_ground[i] = 0.;
  uvec_td = uvec_td_ground + N;  // nodal voltages in time domain
  assert(uvec_td[-1] == 0.);
  vj1 = new double [max_cv];      // control voltages in each time
  in_dj = new double [max_cv*N];
  dj = new double [max_cv];
  
  // Create the identity matrix and its ifft: B=F^(-1)U
  B=new double [fsize*N];
  for (int n(0); n<N; n++) 
    B[n] = 1.;
  for (int m(1); m<num_HB; m++) {
    double step = twopi*m/N;
    double t = 0.;
    int m1 = (2*m-1)*N;
    for (int n(0); n<N; n++) {
      B[m1+n] = cos(t);
      B[m1+N+n] = -sin(t);
      t += step;
    }
  }

  // Linear VCCS
  for (LinVCCSVector::iterator i = linvccs.begin(); i < linvccs.end(); i++){
    double g = (*i)->getGain();
    int Row1 = (*i)->getRow1()*fsize;
    int Row2 = (*i)->getRow2()*fsize;
    int Col1 = (*i)->getCol1()*fsize;
    int Col2 = (*i)->getCol2()*fsize;
    for (int j(0); j<fsize; j++) {
      nam.addElement(Row1+j, Col1+j, g);
      nam.addElement(Row1+j, Col2+j, -g);
      nam.addElement(Row2+j, Col1+j, -g);
      nam.addElement(Row2+j, Col2+j, g);
    }
  }
  // Linear thermal (or convolution) sources. 
  for (LinfVCCSVector::iterator i = linfvccs.begin(); i < linfvccs.end(); i++){
    double_complex y;
    // get DC term
    (*i)->getGain(y, 0.);
    int Row1 = (*i)->getRow1()*fsize;
    int Row2 = (*i)->getRow2()*fsize;
    int Col1 = (*i)->getCol1()*fsize;
    int Col2 = (*i)->getCol2()*fsize;
    // Insert DC component
    double g = y.real();
    addQuad(nam, Row1, Row2, Col1, Col2, g);
    // Insert rest of frequencies
    for (int j(1); j<num_HB; j++) {
      double freq = Omega[j] / twopi;
      (*i)->getGain(y, freq);
      int j1 = 2*j-1;
      addQuad(nam, Row1+j1, Row2+j1, Col1+j1, Col2+j1, y);
    }
  }
  nam.setGMark(); // I think this is right here
  
  // Linear VCQS
  cFactor = 1/dparm[TSTEP];  // 1/h
  for (LinVCQSVector::iterator i=linvcqs.begin(); i<linvcqs.end(); i++){
    double c = (*i)->getGain();
    double geq1 = cFactor * c;
    int Row1 = (*i)->getRow1()*fsize;
    int Row2 = (*i)->getRow2()*fsize;
    int Col1 = (*i)->getCol1()*fsize;
    int Col2 = (*i)->getCol2()*fsize;
    // Add DC component (pure real)
    addQuad(nam, Row1, Row2, Col1, Col2, geq1);
    for (int j(1); j<num_HB; j++) {
      double_complex geq(geq1, c*Omega[j]);
      int j1 = 2*j-1;
      addQuad(nam, Row1+j1, Row2+j1, Col1+j1, Col2+j1, geq);
    }
  }
  nam.setLinMark();
  
  double res_norm(0.);
  int counter(0);
  int iter_acc(0);
  int tot_iter(0);
  int tot_rej(0);
  double Nd = N;
  
  // ---------------------------------------------------------------
  // Begin main time loop
  // ---------------------------------------------------------------
  for (double t(dparm[TSTEP]); t<=dparm[TSTOP]; t+=dparm[TSTEP]) {
    counter++;

    // Clear source vector
    for (int i(0); i<num_HB*nrcs; i++)
      svec[i]=0.;

    // Independent sources
    for (IndepCSVector::iterator i=indcs.begin(); i<indcs.end(); i++) {
      for (int j(0); j<N; j++) 
        (*i) -> eval(put1[j], t, time2[j]);
      int Row1 = (*i)->getRow1()*num_HB;
      int Row2 = (*i)->getRow2()*num_HB;
      fftw_execute(p1);
      put2[0] /= Nd;
      svec[Row1] -= put2[0];
      svec[Row2] += put2[0];
      double tmp = (2./Nd);
      for (int j(1); j<num_HB; j++) {
	put2[j] *= tmp;
	svec[Row1+j] -= put2[j];
	svec[Row2+j] += put2[j];
      }
    }
    
    // Linear capacitor contributions to source vector:
    for (LinVCQSVector::iterator i=linvcqs.begin(); i<linvcqs.end(); i++){
      double geq1= (*i)->getGain() *cFactor;        // c/h
      int Col1 = (*i)->getCol1()*num_HB;
      int Col2 = (*i)->getCol2()*num_HB;
      int Row1 = (*i)->getRow1()*num_HB;
      int Row2 = (*i)->getRow2()*num_HB;
      for (int j(0); j<num_HB; j++){
	double_complex ieq = geq1*(uvec[Col1+j]-uvec[Col2+j]);
        svec[Row1+j] += ieq;
        svec[Row2+j] -= ieq;
      }
    }

    // Time-delayed current sources (this is transitory, only DC
    // component for now).  Add contributions to source vector now,
    // store voltages at the end of Newton iterations.
    for (TDCSVector::iterator i=tdcs.begin(); i<tdcs.end(); i++) {
      // vj1 not used in thermal source, but it will be necessary in a
      // future implementation.
      const int* Col1p = (*i)->getColPlus();
      const int* Col2p = (*i)->getColMinus();
      int ncvolt = (*i)->getNCVolt();
      for (int  k(0); k<ncvolt; k++) 
	vj1[k] = uvec[Col1p[k]*num_HB].real() - uvec[Col2p[k]*num_HB].real();

      double dccurrent;
      (*i) -> eval(dccurrent, vj1);
      int Row1 = (*i)->getRow1()*num_HB;
      int Row2 = (*i)->getRow2()*num_HB;
      svec[Row1] -= dccurrent;
      svec[Row2] += dccurrent;
    }

    // *********************************************************************
    // Solve nonlinear problem
    // *********************************************************************
    int iter = 0;
    int rejections = 0;
    int nlsize = nrcs*fsize;
    
    // Copy nodal voltages and source vectors to double vectors
    c2d(uvec, uvec_d, nrcs, num_HB);
    c2d(svec, svec_d, nrcs, num_HB);
    NewtonSolve<EnvelopeHB> nw(this, &nam, nlsize);
    double residual = nw.solve(uvec_d, svec_d, iter, rejections);
    // Get results to complex format
    d2c(uvec, uvec_d, nrcs, num_HB);    

    // Time-delayed current sources (this is transitory, only DC
    // component for now).  
    // Store voltages at the end of Newton iterations.
    for (TDCSVector::iterator i=tdcs.begin(); i<tdcs.end(); i++) {
      const int* Col1p = (*i)->getColPlus();
      const int* Col2p = (*i)->getColMinus();
      int ncvolt = (*i)->getNCVolt();
      for (int  k(0); k<ncvolt; k++) 
	vj1[k] = uvec[Col1p[k]*num_HB].real() - uvec[Col2p[k]*num_HB].real();
      (*i)->storeVoltages(t, vj1);
    }

    // Calculate voltage values in time domain.  This may not be
    // efficient since possibly not all voltages are needed.
    for (int i(0); i<nrcs; i++){
      denormAndCopy(uvec + i*num_HB, put2, num_HB);
      fftw_execute(p2);
      int ibase = i*N;
      for (int j(0); j<N; j++){
	uvec_td[ibase+j] = put1[j];
      }
    }
    // Calcuate Q(n-1) for nonlinear charges
    for (int i(0); i<nCharges; i++){
      const int* Col1p=genvcqs[i]->getColPlus();
      const int* Col2p=genvcqs[i]->getColMinus();
      for (int j(0); j<N; j++){
	int ncvolt = genvcqs[i]->getNCVolt();
	for (int  k(0); k<ncvolt; k++)
	  vj1[k] = uvec_td[Col1p[k]*N+j] - uvec_td[Col2p[k]*N+j];
	genvcqs[i]->eval(put1[j],vj1);
      }
      fftw_execute(p1);
      normAndCopy(put2, q+i*num_HB, N, num_HB);
    }

    // ************************** Save Results and print status

    // Status
    tot_rej += rejections;
    iter_acc += iter;
    tot_iter += iter;
    if (residual > res_norm)
      res_norm = residual;
    const int& nsteps = getParmSet()->getIHandler()->getParmVec()[OUTSTEPS];
    if (!(counter % nsteps)) {
      int iter_av = iter_acc / nsteps;
      // Print running info
      printf( " %g   \t %d \t\t %d \t\t %g \n", 
	      t, counter, iter_av, res_norm);
      iter_acc = 0;
      res_norm = 0.;
    }

    // Save results
    DVectorHolder* dvh = new DVectorHolder(iv.size(), N, time2);
    // Put time value in main sweep vector
    dvh->setVal(t);
    int iend = iv.size();
    for(int i(0); i<iend; i++) 
      dvh->saveValue(i, uvec_td + iv[i]*N);
    om->addSample(dvh);
  }

  printf("\nTotal (Newton method) iterations: %d\n", tot_iter);  
  printf("\nTotal (Newton method) rejections: %d\n\n", tot_rej);  
  printf("**************** End Of Envelope HB Analysis *******************\n");
  
  delete [] dj;
  delete [] in_dj;
  delete [] vj1;
  delete [] uvec_td_ground;
  delete [] time2;
  delete [] B;
  delete [] q;
  
  fftw_destroy_plan(p2);
  fftw_destroy_plan(p1);
  
  delete [] put2;
  delete [] put1;
  delete [] svec_d;
  delete [] uvec_d;
  delete [] svec_ground;
  delete [] uvec_ground;
  delete [] Omega;
}
  

int EnvelopeHB::fillNAMandRHS(UMFPackDI& nam, 
			      double* svec_r, double* uvec_j_r, 
			      const double& beta, const double& gmin)
{
  int retcode = 1; // 1 means linear problem
  if (genvccs.size() || genvcqs.size())
    retcode = 0; // 0 means normal nonlinear problem

  int fsize = 2*num_HB - 1;
  double Nd = double(N);
  // Use uvec_j to denote previous iteration
  double_complex* uvec_j = uvec;
  // Get vectors to complex format
  d2c(uvec_j, uvec_j_r, nrcs, num_HB);    
  d2c(svec, svec_r, nrcs, num_HB);    

  // Convert previous-iteration voltages to time domain
  // This may not be efficient since not all voltages may be needed.
  for (int i(0); i<nrcs; i++){
    denormAndCopy(uvec_j + i*num_HB, put2, num_HB);
    fftw_execute(p2);
    int ibase = i*N;
    for (int j(0); j<N; j++){
      uvec_td[ibase+j]=put1[j];
    }
  }
  // Nonlinear VCCS ----------------------------------------------
  for (unsigned i(0); i<genvccs.size(); i++){            
    const int* Col1p = genvccs[i]->getColPlus();
    const int* Col2p = genvccs[i]->getColMinus();
    int ncvolt = genvccs[i]->getNCVolt();
    // get ij, dj, dj*vc
    for (int j(0); j<N; j++){
      for (int  k(0); k<ncvolt; k++) 
	vj1[k] = uvec_td[Col1p[k]*N+j] - uvec_td[Col2p[k]*N+j];
      genvccs[i]->eval_and_deriv(put1[j], dj, vj1);
      for (int  k(0); k<ncvolt; k++){
	in_dj[k*N+j] = dj[k] * beta;
	// Add contribution from Jacobian matrix to source vector
	put1[j] -= dj[k]*vj1[k] * beta;
      }
    }
    //fftw3: ij -> Ij
    int Row1=genvccs[i]->getRow1()*num_HB;
    int Row2=genvccs[i]->getRow2()*num_HB;
    fftw_execute(p1);
    put2[0] /= Nd;
    svec[Row1] -= put2[0];
    svec[Row2] += put2[0];
    double tmp = (2./Nd);
    for (int j(1); j<num_HB; j++) {
      put2[j] *= tmp;
      svec[Row1+j] -= put2[j];
      svec[Row2+j] += put2[j];
    }
    //fill NAM
    Row1 = genvccs[i]->getRow1()*fsize;
    Row2 = genvccs[i]->getRow2()*fsize;
    for (int  k(0); k<ncvolt; k++) {
      int Col1 = Col1p[k]*fsize;
      int Col2 = Col2p[k]*fsize;
      // DC column (m=0) treated specially
      for (int n(0); n<N; n++)
	put1[n] = (in_dj[k*N+n]+gmin)*B[n];
      fftw_execute(p1);
      // DC part
      put2[0] /= Nd;
      addQuad(nam, Row1, Row2, Col1, Col2, put2->real());
      // All other rows 
      double tmp(2./Nd);
      for (int j(1); j<num_HB; j++) {
	put2[j] *= tmp;
	int j1 = 2*j-1;
	addQuad(nam, Row1+j1, Row2+j1, Col1, Col2, put2[j].real());
	j1++;
	addQuad(nam, Row1+j1, Row2+j1, Col1, Col2, put2[j].imag());
      }
      // All other columns
      for (int m(1); m<num_HB; m++){
	int m1 = 2*m-1;
	for (int n(0); n<N; n++)
	  put1[n] = (in_dj[k*N+n]+gmin)*B[m1*N+n];
	fftw_execute(p1);
	// DC part
	put2[0] /= Nd;
  	addQuad(nam, Row1, Row2, Col1+m1, Col2+m1, put2->real());
	// All other rows
	for (int j(1); j<num_HB; j++) {
	  put2[j] *= tmp;
	  int j1 = 2*j-1;
	  addQuad(nam, Row1+j1, Row2+j1, Col1+m1, Col2+m1, put2[j].real());
	  j1++;
	  addQuad(nam, Row1+j1, Row2+j1, Col1+m1, Col2+m1, put2[j].imag());
	}
	m1++;
	for (int n(0); n<N; n++)
	  put1[n] = (in_dj[k*N+n]+gmin)*B[m1*N+n];
	fftw_execute(p1);
	// DC part
	put2[0] /= Nd;
  	addQuad(nam, Row1, Row2, Col1+m1, Col2+m1, put2->real());
	// All other rows
	for (int j(1); j<num_HB; j++) {
	  put2[j] *= tmp;
	  int j1 = 2*j-1;
	  addQuad(nam, Row1+j1, Row2+j1, Col1+m1, Col2+m1, put2[j].real());
	  j1++;
	  addQuad(nam, Row1+j1, Row2+j1, Col1+m1, Col2+m1, put2[j].imag());
	}
      }
    }
  }

  // Nonlinear charge sources (VCQS) ------------------------------
  for (int i(0); i<nCharges; i++){
    const int* Col1p=genvcqs[i]->getColPlus();
    const int* Col2p=genvcqs[i]->getColMinus();
    int ncvolt = genvcqs[i]->getNCVolt();
    for (int j(0); j<N; j++){
      for (int  k(0); k<ncvolt; k++)
	vj1[k] = uvec_td[Col1p[k]*N+j] - uvec_td[Col2p[k]*N+j];
      genvcqs[i]->eval_and_deriv(put1[j], dj, vj1);
      for (int k(0); k<ncvolt; k++){
	in_dj[k*N+j] = dj[k] * beta;
	put1[j] -= dj[k]*vj1[k] * beta;
      }
    }
    int Row1=genvcqs[i]->getRow1()*num_HB;
    int Row2=genvcqs[i]->getRow2()*num_HB;
    fftw_execute(p1);
    put2[0] /= Nd;
    put2[0] -= q[i*num_HB];
    put2[0] *= cFactor;
    svec[Row1] -= put2[0];
    svec[Row2] += put2[0];
    double tmp = (2./Nd);
    for (int j(1); j<num_HB; j++) {
      put2[j] *= tmp;
      put2[j] *= double_complex(cFactor, Omega[j]);
      // Check the following!!!!!!!!!!!!!!!!!!!!!!!!
      put2[j] -= q[i*num_HB+j] * cFactor;
      svec[Row1+j] -= put2[j];
      svec[Row2+j] += put2[j];
    }

    //fill NAM
    Row1 = genvcqs[i]->getRow1()*fsize;
    Row2 = genvcqs[i]->getRow2()*fsize;
    for (int  k(0); k<ncvolt; k++) { 
      int Col1 = Col1p[k]*fsize;
      int Col2 = Col2p[k]*fsize;
      // DC column (m=0) treated specially
      for (int n(0); n<N; n++)
	put1[n] = in_dj[k*N+n]*B[n];
      fftw_execute(p1);
      // DC part
      put2[0] *= cFactor/Nd;
      addQuad(nam, Row1, Row2, Col1, Col2, put2->real());
      // All other rows 
      double tmp(2./Nd);
      for (int j(1); j<num_HB; j++) {
	put2[j] *= tmp;
	put2[j] *= double_complex(cFactor, Omega[j]);
	int j1 = 2*j-1;
	addQuad(nam, Row1+j1, Row2+j1, Col1, Col2, put2[j].real());
	j1++;
	addQuad(nam, Row1+j1, Row2+j1, Col1, Col2, put2[j].imag());
      }
      // All other columns
      for (int m(1); m<num_HB; m++){
	int m1 = 2*m-1;
	for (int n(0); n<N; n++)
	  put1[n] = in_dj[k*N+n]*B[m1*N+n];
	fftw_execute(p1);
	// DC part
	put2[0] *= cFactor/Nd;
  	addQuad(nam, Row1, Row2, Col1+m1, Col2+m1, put2->real());
	// All other rows
	for (int j(1); j<num_HB; j++) {
	  put2[j] *= tmp;
	  put2[j] *= double_complex(cFactor, Omega[j]);
	  int j1 = 2*j-1;
	  addQuad(nam, Row1+j1, Row2+j1, Col1+m1, Col2+m1, put2[j].real());
	  j1++;
	  addQuad(nam, Row1+j1, Row2+j1, Col1+m1, Col2+m1, put2[j].imag());
	}
	m1++;
	for (int n(0); n<N; n++)
	  put1[n] = in_dj[k*N+n]*B[m1*N+n];
	fftw_execute(p1);
	// DC part
	put2[0] *= cFactor/Nd;
  	addQuad(nam, Row1, Row2, Col1+m1, Col2+m1, put2->real());
	// All other rows
	for (int j(1); j<num_HB; j++) {
	  put2[j] *= tmp;
	  put2[j] *= double_complex(cFactor, Omega[j]);
	  int j1 = 2*j-1;
	  addQuad(nam, Row1+j1, Row2+j1, Col1+m1, Col2+m1, put2[j].real());
	  j1++;
	  addQuad(nam, Row1+j1, Row2+j1, Col1+m1, Col2+m1, put2[j].imag());
	}
      }
    }
  }

  // Copy source vector to double vector
  c2d(svec, svec_r, nrcs, num_HB);

  return retcode;
}


void EnvelopeHB::normAndCopy(double_complex* source, 
			     double_complex* dest1,
			     const int& nsamples, const int& nh)
{
  *dest1 = (*source)/double(nsamples);
  double tmp(2./double(nsamples));
  double_complex* pend = dest1 + nh;
  for (source++, dest1++; dest1 < pend; source++, dest1++)
    (*dest1) = (*source) * tmp;
}

void EnvelopeHB::denormAndCopy(double_complex* source, 
			       double_complex* dest,
			       const int& nh)
{
  dest[0] = source[0];
  for (int i(1); i<nh; i++)
    dest[i] = source[i] * .5;
  // Erase extra frequencies (due to oversampling)
  for (int i(nh); i<padsize; i++)
    dest[i] = 0.;
}

void EnvelopeHB::addQuad(UMFPackDI& nam, 
			 int row1, int row2,
			 int col1, int col2,
			 const double& g)
{
  nam.addElement(row1, col1, g);
  nam.addElement(row1, col2, -g);
  nam.addElement(row2, col1, -g);
  nam.addElement(row2, col2, g);
}

void EnvelopeHB::addQuad(UMFPackDI& nam, 
			 int row1, int row2,
			 int col1, int col2,
			 const double_complex& g)
{
  // Change to double
  const double& greal = g.real();
  const double& gimag = g.imag();
  // 0,0: a
  nam.addElement(row1, col1, greal);
  nam.addElement(row1, col2, -greal);
  nam.addElement(row2, col1, -greal);
  nam.addElement(row2, col2, greal);
  // 0,1: -b
  nam.addElement(row1, col1+1, -gimag);
  nam.addElement(row1, col2+1, gimag);
  nam.addElement(row2, col1+1, gimag);
  nam.addElement(row2, col2+1, -gimag);
  // 1,0: b
  nam.addElement(row1+1, col1, gimag);
  nam.addElement(row1+1, col2, -gimag);
  nam.addElement(row2+1, col1, -gimag);
  nam.addElement(row2+1, col2, gimag);
  // 1,1: a
  nam.addElement(row1+1, col1+1, greal);
  nam.addElement(row1+1, col2+1, -greal);
  nam.addElement(row2+1, col1+1, -greal);
  nam.addElement(row2+1, col2+1, greal);
}

void EnvelopeHB::c2d(double_complex* uvec, double* uvec_d, 
		     const int& nrcs, const int& num_HB)
{
  int fsize = 2*num_HB - 1;
  // DC component first
  for (int i(0); i<nrcs; i++) {
    uvec_d[i*fsize] = uvec[i*num_HB].real();
  }
  // All other harmonics
  for (int i(0); i<nrcs; i++) {
    int id = i*fsize + 1;
    int ic = i*num_HB + 1;
    for (int j(1); j<num_HB; j++) {
      uvec_d[id] = uvec[ic].real();
      uvec_d[id+1] = uvec[ic].imag();
      id += 2;
      ic++; 
    }
  }
}


void EnvelopeHB::d2c(double_complex* uvec, double* uvec_d, 
		     const int& nrcs, const int& num_HB)
{
  int fsize = 2*num_HB - 1;
  // DC component first
  for (int i(0); i<nrcs; i++) {
    uvec[i*num_HB] = uvec_d[i*fsize];
  }
  // All other harmonics
  for (int i(0); i<nrcs; i++) {
    int id = i*fsize + 1;
    int ic = i*num_HB + 1;
    for (int j(1); j<num_HB; j++) {
      uvec[ic] = double_complex(uvec_d[id], uvec_d[id+1]); 
	// *(reinterpret_cast<double_complex*>(uvec_d + id));
      id += 2;
      ic++; 
    }
  }
}
