//
// Convergence seems to be reasonable but many iterations are usually
// required if the initial guess is bad.
//
// by Carlos E. Christoffersen

#include <carrot/Circuit.h>
#include <carrot/OP.h>
#include <carrot/Options.h>
#include "OPA.h"
#include <carrot/UMFPackDI.h>
#include "global.h"

// For now put the include here to test the routines
#include <carrot/DoubleHolder.h>

#include <stdlib.h>
#include <cmath>
#include <string>
#include <iostream>

// Debug
// #define DEBUG 1

#ifdef DEBUG
#include <cstdio>
#endif

static ItemInfo OPA_info = {
  "op",
  "Operating Point Analysis",
  "Carlos Christoffersen",
  DEFAULT_ADDRESS
};

enum {MAX_DPARM_INDEX}; 

// static ParamHandler<double>::ParmInfo OPA_dparams[] = {
//   {"ss_inc", "Source-stepping increment factor", false, 1.3}
// };

// ***********************************************************************
// Integer parameters
// ***********************************************************************
enum {SENS, NORM, DETAIL, MAX_IPARM_INDEX};

static ParamHandler<int>::ParmInfo OPA_iparams[] = {
  {"sens", "Perform Sensitivity analysis (0: no)", false, 0},
  {"norm", "Calculate normalised sensitivities if possible (0: no)", false, 1},
  {"detail", "Print detailed OP information (long, 0:no)", false, 0}
};

// ***********************************************************************
// String parameters
// ***********************************************************************
enum {NODENAME, MODELNAME, MAX_SPARM_INDEX};

static ParamHandler<std::string>::ParmInfo OPA_sparams[] = {
  {"nodename", "Name of the node to calculate sensitivities", false, "out1"},
  {"modelname", "Sensitivities are calculated with respect to this model", 
   false, "all"}
};

extern "C" {

  //Export the name of the module we are using
  const char * AnalysisName(){
    return OPA_info.type.c_str();
  }
  ParmSet * newParmSet(const char * model){
    return new ParmSet(model, &OPA_info, 
		       NULL, MAX_DPARM_INDEX, 
		       OPA_iparams, MAX_IPARM_INDEX,
		       OPA_sparams, MAX_SPARM_INDEX);
  }
  void delParmSet(ParmSet * p){
    delete p;
    p = 0;
  }
  Analysis * newObject(ParmSet * p){
    return new OPA(p);
  }
  void delObject(Analysis * e){
    delete e;
    e = 0;
  }
};

void OPA::run(Circuit* cir, OutputManager* om)
{
  // Check parameters
  getParmSet()->checkParms();

  // Set variables in OutputManager
  om->setVarUnit("-"); // For now no units if no sweep
  om->setAnalysis("op");

  //  double *dparm = getParmSet()->getDHandler()->getParmVec();
  int *iparm = getParmSet()->getIHandler()->getParmVec();
  std::string *sparm = getParmSet()->getSHandler()->getParmVec();

  printf("****************************************************************\n");
  printf("*                  Operating Point Analysis                    *\n");
  printf("****************************************************************\n");

  // Set row/column numbers in terminals and get number of OP-used
  // terminals. Linear and nonlinear charge sources are excluded to
  // avoid a singular matrix when a node is not resistively connected
  // to ground. In this pathological case the node numbering is not
  // the same as when all elements are included.
  int nterms = cir->setMatrixRCs();
  // Get rows of desired output variables
  std::vector<int> iv;
  om->getRowNumbers(cir, iv);
  // ***********************  Create matrix and vectors

  // Ground node not included in nrcs
  int nrcs = nterms - 1; 
  // NAM matrix. No need to clean matrix because it is fresh.
  UMFPackDI nam(nrcs, 24*nrcs); 

  // Right hand side vector (source vector)
  double *svec_ground = new double[nrcs+1]; 
  double *svec = svec_ground + 1;
  // unknowns vector (with ground)
  double *uvec_ground = new double[nrcs+1]; 
  // Ground node potential is zero
  uvec_ground[0] = 0.; 
  // This would be uvec_(j+1)
  double *uvec = uvec_ground + 1;

  // Set initial guess to zero
  for (int i(0); i<nrcs; i++)
    uvec[i] = 0.;

  // For now just calculate one point
  OP opoint(cir);
  int iter, rejections;
  double residual = opoint.op(nrcs, nam, uvec, svec, iter, rejections);

  // ****************   Save results here 
  // om takes care of cleaning this ...
  DoubleHolder* dh = new DoubleHolder(iv.size()); 
  // For now add a zero in value (we'll figure this out later)
  dh->setVal(0.);
  int iend = iv.size();
  for(int i(0); i<iend; i++) 
    dh->saveValue(i, uvec[iv[i]]);
  om->addSample(dh);

  printf("\nIterations = %d\n", iter);
  printf("Rejections = %d\n", rejections);
  printf("Residual = %g\n\n", residual);

  if (residual > options.getDHandler()->getParmVec()[OPT_TOL])
    printf("WARNING: Convergence not achieved. Results possibly wrong.\n\n");

  if (iparm[DETAIL]) {
    printf("****************************************************************\n");
    printf("*            Detailed operating point information              *\n");
    printf("****************************************************************\n");
    // Get the maximum number of controlling voltages we can expect
    const int& max_cv = cir->getMaxCV();
    double* vj = new double[max_cv];
    double* dij = new double[max_cv];
    MacroElVector mev = cir->getMElementvec();
    // Retrieve and print information about each macroelement
    for (MacroElVector::iterator j = mev.begin(); j < mev.end(); j++) {
      MacroElement *me = *j;
      // Print MacroElement information
      printf("----------------------------------------------------------------------\n");
      printf("  Device: %s\n\n", me->getName().c_str());
      
      // Independent sources
      IndepCSVector indv = me->getIndepCSvec();
      if (indv.size())
	printf("\tIndependent sources:\n");
      for (IndepCSVector::iterator i = indv.begin(); i < indv.end(); i++) {
	// Get values (there are no derivatives)
	double ivalue;
	(*i) -> eval(ivalue, 0.);
	printf("\t\tivalue = %g\n", ivalue);
      }
      // No point printing information about Linear CS
      // No point printing information about Linear QS
      
      // Generic VCCS
      GenVCCSVector gencv = me->getGenVCCSvec();
      if (gencv.size())
	printf("\tNonlinear current sources:\n");
      for (GenVCCSVector::iterator i = gencv.begin(); i < gencv.end(); i++) {
	// Calculate vj.
	for (int k(0); k < (*i)->getNCVolt(); k++) {
	  // uvec[-1] = 0 (takes care of ground node)
	  vj[k] = uvec[(*i)->getColPlus()[k]]
		- uvec[(*i)->getColMinus()[k]];
	}
	// Get values and derivatives
	double ij;
	(*i)->eval_and_deriv(ij, dij, vj);
	printf("\t\ti = %g\n", ij);
	for (int k(0); k < (*i)->getNCVolt(); k++) {
	  printf("\t\tv%d = %g  \tdi/dv%d = %g\n",k , vj[k], k, dij[k]);
	}
	printf("\n");
      }
      
      // Generic VCQS
      GenVCQSVector genqv = me->getGenVCQSvec();
      if (genqv.size())
	printf("\tNonlinear charge sources:\n");
      for (GenVCQSVector::iterator i = genqv.begin(); i < genqv.end(); i++) {
	// Calculate vj.
	for (int k(0); k < (*i)->getNCVolt(); k++) {
	  // uvec[-1] = 0 (takes care of ground node)
	  vj[k] = uvec[(*i)->getColPlus()[k]]
		- uvec[(*i)->getColMinus()[k]];
	}
	// Get values and derivatives
	double qj;
	(*i)->eval_and_deriv(qj, dij, vj);
	printf("\t\tq = %g\n", qj);
	for (int k(0); k < (*i)->getNCVolt(); k++) {
	  printf("\t\tv%d = %g  \tdq/dv%d = %g\n",k , vj[k], k, dij[k]);
	}
	printf("\n");
      }
    }
    delete [] dij;
    delete [] vj;
    printf("----------------------------------------------------------------------\n");
    // That is all for now. Remember to add new types later.
  }

  // For now, it is better to do things here because we already have
  // all vectors and matrices set. Later we may think some way to move
  // this to a separate (hopefully general) routine.
  if (iparm[SENS]) {
    printf("****************************************************************\n");
    printf("*                   OP Sensitivity Analysis                    *\n");
    printf("****************************************************************\n");
    // We need to know the sensitivity of what function of voltage we
    // calculate (phi function). 
    // Search the node number.
    Terminal* term = cir->getTerminal(sparm[NODENAME]);
    // Make sure the terminal exists
    // Good place for exception throwing
    if (term == NULL) {
      fprintf(stderr, "Node \"%s\" could not be found.\n\n", 
	      sparm[NODENAME].c_str());
      exit(0);
    }
    int vnumber = term->getRC();
    printf("\n *** Sensitivities of potential at node %s \n",
	   term->getName().c_str());

    double *d = new double[nrcs];
    for (int i(0); i < nrcs; i++) 
      d[i] = 0.;
    d[vnumber] = 1.;
    // Adjoint voltages vector
    double *uvec_a = new double[nrcs];
    // Right-hand-side derivative vector
    double *rhs_deriv_ground = new double[nrcs+1];
    //    rhs_deriv_ground[0] = 0.; not needed
    double *rhs_deriv = rhs_deriv_ground + 1;
    // Solve adjoint system
    nam.solveTransposed(uvec_a, d);

#ifdef DEBUG
    nam.printTriplet();
    printf("index: %d\n", vnumber);
    for (int i(0); i < cir_nterms; i++) {
      int row = termvec[i]->getRC();
      if (row != -1)
	printf("va( %s ) = %g V, \t\t %d\n", 
	       termvec[i]->getName().c_str(), uvec_a[row], row);
    }
#endif
    
    // Get the maximum number of controlling voltages we can expect
    const int& max_cv = cir->getMaxCV();
    double* vj = new double[max_cv];
    // Go for each parameter in each ParmSet
    int max_parms = cir->getMaxDParms();
    int max_CS = cir->getMaxCS();
    // Jac: row-oriented matrix. Each row contains the Jacobian for
    // one current source in a MacroElement.
    double *Jac = new double[max_parms * max_CS];
    double *sens = new double[max_parms];

    if (sparm[MODELNAME] == "all") {
      // Do sensitivities with respect to all parameters
      printf("     with respect to all parameters requested.\n");

      ParmSetMap psetmap = cir->getParmSetMap();
      ParmSetMap::iterator psetend = psetmap.end();
      for (ParmSetMap::iterator i = psetmap.begin(); i != psetend; i++) {
	// If a ParmSet has no double parameters (getDHandler returns
	// NULL) then we calculate no derivatives:
	if (i->second->getDHandler() == NULL)
	  continue;
	getSens(i->second, term, nrcs, max_parms, max_CS, 
		uvec, vj, d, uvec_a, rhs_deriv, Jac, sens);
      }
    } else {
      printf("     with respect to parameters in model: %s requested.\n",
	     sparm[MODELNAME].c_str());
      ParmSet *pset = cir->getParmSet(sparm[MODELNAME]);
      if (pset == NULL) {
	fprintf(stderr, "Model \"%s\" could not be found.\n\n", 
		sparm[MODELNAME].c_str());
	exit(0);
      }
      if (pset->getDHandler() == NULL) {
	fprintf(stderr, "Model \"%s\" has no double parameters.\n\n", 
		sparm[MODELNAME].c_str());
	exit(0);
      }
      getSens(pset, term, nrcs, max_parms, max_CS, 
	      uvec, vj, d, uvec_a, rhs_deriv, Jac, sens);

    }
    // Free stuff allocated here
    delete [] sens;
    delete [] Jac;
    delete [] rhs_deriv_ground;
    delete [] uvec_a;
    delete [] d;
  }

  // Clean up everything (in reverse order they were created)
  delete [] uvec_ground;
  delete [] svec_ground;

  // nam gets destroyed as it goes out of scope
  printf("********************** End Of OP Analysis **********************\n");
}


void OPA::getSens(ParmSet* ps, Terminal *term,
		 const int& nrcs, const int& max_parms, const int& max_CS, 
		 const double *uvec, double *vj,
		 const double *d, const double *uvec_a, double *rhs_deriv,
		 double *Jac, double *sens)
{
  int *iparm = getParmSet()->getIHandler()->getParmVec();
  int vnumber = term->getRC();

  unsigned nparms = ps->getDHandler()->getNParms();
  // Erase sensitivity vector (new parameter set)
  for (unsigned j(0); j < nparms; j++) 
    sens[j] = 0.;
  // Go for each MacroElement that uses ps
  // Calculate all derivatives in this MacroElement first
  bool calculated = false;
  unsigned nme = ps->getMEVecSize();
  for (unsigned j(0); j < nme; j++) {
    MacroElement *me = ps->getME(j);
    // Independent sources
    IndepCSVector ind1 = me->getIndepCSvec();
    // We do not need to calculate this again for all
    // MacroElements with the same parameter set;
    if (!calculated)
      for (unsigned k(0); k < ind1.size(); k++) {
	// Calculate derivatives 
	bool flag = ind1[k]->paramDeriv(Jac + k*nparms, 0.);
	if (!flag) {
#ifdef DEBUG
	  printf("\nWarning: (ind) paramDeriv() not implemented in model: %s\n\n", 
		 ps->getName().c_str());
#endif
	  // Derivatives not implemented, set Jacobian row to zero
	  double *max_p = Jac + (k+1)*nparms;
	  for (double* jac_p(Jac + k*nparms); jac_p < max_p; jac_p++)
	    (*jac_p) = 0.;
	}
      }
    unsigned kbase = ind1.size();
    // Linear CS
    LinVCCSVector lin1 = me->getLinearVCCSvec();
    if (!calculated) {
      for (unsigned k(0); k < lin1.size(); k++) {
	bool flag = lin1[k]->paramDeriv(Jac + (kbase+k) * nparms);
	if (!flag) {
#ifdef DEBUG
	  printf("\nWarning: (lin) paramDeriv() not implemented in model: %s\n\n", 
		 ps->getName().c_str());
#endif
	  // Derivatives not implemented, set Jacobian row to zero
	  double *max_p = Jac + (kbase+k+1)*nparms;
	  for (double* jac_p(Jac+(kbase+k)*nparms); jac_p < max_p; jac_p++)
	    (*jac_p) = 0.;
	}
      }
      calculated = true;
    }
    kbase += lin1.size();
    // Generic VCCS
    GenVCCSVector gen1 = me->getGenVCCSvec();
    for (unsigned k(0); k < gen1.size(); k++) {
      // Calculate voltage vector (vj) for VCCS
      for (int l(0); l < gen1[k]->getNCVolt(); l++) {
	// uvec[-1] = 0 (takes care of ground node)
	vj[l] = uvec[gen1[k]->getColPlus()[l]] 
	  - uvec[gen1[k]->getColMinus()[l]];
      }
      // Calculate derivatives 
      bool flag = gen1[k]->paramDeriv(Jac + (kbase+k)*nparms, vj);
      if (!flag) {
#ifdef DEBUG
	printf("\nWarning: (gen) paramDeriv() not implemented in model: %s\n\n", 
	       ps->getName().c_str());
#endif
	// Derivatives not implemented, set Jacobian row to zero
	double *max_p = Jac + (kbase+k+1)*nparms;
	for (double* jac_p(Jac+(kbase+k)*nparms); jac_p < max_p; jac_p++)
	  (*jac_p) = 0.;
      }
    }
    // Now derivatives (with respect to all parameters are stored
    // in Jac. Calculate contributions to rhs_deriv for each parameter.
    for (unsigned k(0); k < nparms; k++) {
      // Erase vector to calculate new derivatives
      for (int l(-1); l < nrcs; l++)
	rhs_deriv[l] = 0.;
      for (unsigned l(0); l < ind1.size(); l++) {
	// Store derivatives.
	// *** IMPORTANT: Here I have changed the signs to get the
	// correct result. Make sure to verify this!!!
	rhs_deriv[ind1[l]->getRow1()] -= Jac[l*nparms + k];
	rhs_deriv[ind1[l]->getRow2()] += Jac[l*nparms + k];
      }
      unsigned lbase = ind1.size();
      for (unsigned l(0); l < lin1.size(); l++) {
	// Store derivatives
	double deriv = Jac[(lbase+l)*nparms + k] 
	  * (uvec[lin1[l]->getCol1()] - uvec[lin1[l]->getCol2()]);
	rhs_deriv[lin1[l]->getRow1()] -= deriv;
	rhs_deriv[lin1[l]->getRow2()] += deriv;
      }
      lbase += lin1.size();	  
      for (unsigned l(0); l < gen1.size(); l++) {
	// Store derivatives
	rhs_deriv[gen1[l]->getRow1()] -= Jac[(lbase+l)*nparms + k];
	rhs_deriv[gen1[l]->getRow2()] += Jac[(lbase+l)*nparms + k];
      }
      // Now multiply: uvec_a' * rhs_deriv
      for (int l(0); l < nrcs; l++) {
	// Add contributions of all MacroElements using the same
	// parameter set.
	sens[k] += uvec_a[l] * rhs_deriv[l];
      }
    }
  }
  // Print sensitivities
  printf("\n  +-----------------------------------------------------+\n");
  printf("  | Model: %16s                             |\n",
	 ps->getName().c_str());
  printf("  +-----------------------------------------------------+\n");
  if (iparm[NORM]) {
    printf("  | Parameter Name    | Value          | Sensitivity    |\n");
  } else {
    printf("  | Parameter Name    | Value          | Derivative     |\n");
  }
  printf("  +-----------------------------------------------------+\n");
  double *curr_dparm = ps->getDHandler()->getParmVec();
  for (unsigned k(0); k<nparms; k++) {
    if (sens[k] != 0) {
      // To avoid printing unnecessary lines omit printing of zero
      // sensitivities.
      if (iparm[NORM]) {
	double nsens = sens[k];
	if (curr_dparm[k] != 0.) {
	  // Normalise sensitivity
	  nsens *= curr_dparm[k];
	} 
	if (uvec[vnumber] != 0.) {
	  // Normalise sensitivity
	  nsens /= uvec[vnumber];
	} 
	printf("  | %16s  | %+.5e   | %+.5e   |\n", 
	       ps->getDHandler()->getParmDesc(k).name.c_str(), 
	       curr_dparm[k], nsens);
      } else {
	// Only double parameters matter
	printf("  | %16s  | %+.5e   | %+.5e   |\n", 
	       ps->getDHandler()->getParmDesc(k).name.c_str(), 
	       curr_dparm[k], sens[k]);
      }
    }
  }
  printf("  +-----------------------------------------------------+\n\n");

}
