// This may look like C code, but it is really -*- C++ -*-
//
// Implements BE integration method
//
// 
// by Carlos E. Christoffersen

#ifndef BEuler_H
#define BEuler_H
#include <cmath>

class BEuler
{
public:

  BEuler(const int& nrcs, const int& ncharges) 
    : nrcs(nrcs), ncharges(ncharges),
      tstep(0.), tstep1(0.), inv_tstep(0.)
  {
    double *uvecn_ground = new double[nrcs + 1];
    uvecn_ground[0] = 0.;
    uvecn = uvecn_ground + 1;
    qn = new double[ncharges];
    uvecn1 = new double[nrcs];
  }

  ~BEuler() 
  {
    delete [] uvecn1;
    delete [] qn;
    delete [] (uvecn - 1);
  }

  // Initialise object
  void init(const double* uvec, const double* q, const double& h)
  {
    tstep1 = tstep = h;
    for (int i=0; i<nrcs; i++) 
      uvecn1[i] = uvecn[i] = uvec[i];
    for (int i=0; i<ncharges; i++)
      qn[i] = q[i];
  }

  // Set time step value
  inline void seth(const double& h)
  {
    tstep = h;
    inv_tstep = 1. / h;
  }

  // Returns the factor to multiply the C matrix.
  inline const double& getCfactor()
  {
    return inv_tstep;
  }

  // Returns the current contribution of a linear capacitance
  inline double getCieq(const int& col1, const int& col2, const double& c)
  {
    return (uvecn[col1] - uvecn[col2]) * c * inv_tstep;
  }

  // Returns charge contribution only
  inline void getIeqQ(const double& qj, const int& i,
		      double& ieq)
  {
    ieq = (qn[i] - qj) * inv_tstep;
  }

  // Returns ieq and geq for each controlling voltage
  inline void getIeqGeq(const double& dqj, 
			const double& vj, 
			double& ieq, double& geq)
  {
    ieq = (dqj * vj) * inv_tstep;
    geq = dqj * inv_tstep;
  }

  // Save previous values for next time step
  void advance(const double* uvec, const double* q)
  {
    tstep1 = tstep;
    for (int i=0; i<nrcs; i++) {
      uvecn1[i] = uvecn[i];
      uvecn[i] = uvec[i];
    }
    for (int i=0; i<ncharges; i++)
      qn[i] = q[i];
  }

  // Calculate truncation error (delta)
  inline void getTruncError(const double* uvec, 
		double& delta, double& newtstep, double& maxdelta)
  {
    delta  = 0.;
    double truncError;
    for( int i(0); i < nrcs; i++ ) {
      truncError = fabs(uvec[i] - uvecn[i] 
			- (uvecn[i] - uvecn1[i]) * tstep / tstep1);
      // Make a relative error (if possible)
      double v = fabs(uvecn[i]);
      if (v > 1.)
	truncError /= v;
      if ( truncError > delta )
	delta = truncError;
    } 
    newtstep = tstep * maxdelta / delta;
  }

private:
  
  int nrcs;
  int ncharges;
  double tstep, tstep1;
  double inv_tstep;
  double* uvecn;
  double* uvecn1;
  double* qn;

};

#endif
