// This may look like C code, but it is really -*- C++ -*-
//
// Regular transient analysis
//
// by Jahanzeb Tariq Khan and Carlos Christoffersen

#include <carrot/Circuit.h>
#include <carrot/OP.h>

#include <stdlib.h>
#include <cmath>
#include <string>
#include <iostream>

// Debug
//#define DEBUG 1

#include <cstdio>

static ItemInfo Transient_info = {
  "tran",
  "Transient Analysis",
  "Jahanzeb Khan, Carlos Christoffersen",
  DEFAULT_ADDRESS
};

// ***********************************************************************
// Double parameters
// ***********************************************************************
enum {TSTEP, TSTOP, MINTSTEP, MAXDELTA, TSAVE, MAX_DPARM_INDEX};
static ParamHandler<double>::ParmInfo Transient_dparams[] = {
  {"tstep", "Time step value: h (s)", true, 0.},
  {"tstop", "Stop time (s)", true, 0.},
  {"mintstep", "Minimum Time Step", false, 1e-18},
  {"maxdelta", "Maximum Value of Delta", false, 0.5},
  {"tsave", "Time to start saving (s)", false, 0.}
};

// ***********************************************************************
// String Parameters
// ***********************************************************************
// enum {NODENAME, MAX_SPARM_INDEX };
// static ParamHandler<std::string>::ParmInfo Transient_sparams[] = {
//   {"nodename", "Name of the node to calculate sensitivities", false, "1"}
// };

// ***********************************************************************
// Integer parameters
// ***********************************************************************
enum {SENS, NORM, IM, DCOP, VARTSTEP, OUTSTEP, MAX_IPARM_INDEX};
static ParamHandler<int>::ParmInfo Transient_iparams[] = {
  {"sens", "Perform Sensitivity analysis (0: no)", false, 0},
  {"norm", "Calculate normalised sensitivities if possible (0: no)", false, 1},
  {"im", "Number of the integration method. B.E.=0, Trap=1 ", false, 0 },
  {"dcop", "Option to start with DC Operating Point. 0=F, 1=T ", false, 1 },
  {"vartstep", "Option to use variable timestep. 0=F, 1=T", false, 1 },
  {"outsteps", "Number of steps to show simulation progress", false, 100 }
};

#include "Transient.h"
#include "global.h"
#include "BEuler.h"
#include "Trapezoidal.h"
#include <carrot/DoubleHolder.h>

extern "C" {

  //Export the name of the module we are using
  const char * AnalysisName(){
    return Transient_info.type.c_str();
  }
  ParmSet * newParmSet(const char * model){
    return new ParmSet(model, &Transient_info,
		       Transient_dparams, MAX_DPARM_INDEX,
		       Transient_iparams, MAX_IPARM_INDEX);
  }
  void delParmSet(ParmSet * p){
    delete p;
    p = 0;
  }
  Analysis * newObject(ParmSet * p){
    return new Transient(p);
  }
  void delObject(Analysis * e){
    delete e;
    e = 0;
  }
};

void Transient::saveVector(const double& t, double* uvec,
			   const int& iter, const double& residual)
{
  counter++;
  iter_acc += iter;
  if (residual > res_norm)
    res_norm = residual;
  const int& nsteps = getParmSet()->getIHandler()->getParmVec()[OUTSTEP];
  if (!(counter % nsteps)) {
    int iter_av = iter_acc / nsteps;
    // Print running info
    printf( " %g \t %d \t\t %d \t\t %g \n", t, counter, iter_av, res_norm);
    iter_acc = 0;
    res_norm = 0.;
  }

  // ************************** Save results
  const double& tsave = getParmSet()->getDHandler()->getParmVec()[TSAVE];
  if (t >= tsave) {
    DoubleHolder* dh = new DoubleHolder(iv.size());
    // Put time value in main sweep vector
    dh->setVal(t);
    int iend = iv.size();
    for(int i(0); i<iend; i++) 
      dh->saveValue(i, uvec[iv[i]]);
    om->addSample(dh);
  }
}

void Transient::run(Circuit* cir, OutputManager* om)
{
  // Setup output manager main paramenters
  this->om = om;
  om->setVarUnit("s"); // Seconds
  om->setAnalysis("transient");
  // Reset variables for output routine
  counter = 0; // reset step counter
  iter_acc = 0;
  res_norm = 0.;

  // Check parameters
  getParmSet()->checkParms();

  double *dparm = getParmSet()->getDHandler()->getParmVec();
  int *iparm = getParmSet()->getIHandler()->getParmVec();
  // std::string *sparm = getParmSet()->getSHandler()->getParmVec();

  printf("****************************************************************\n");
  printf("*                     Transient Analysis                       *\n");
  printf("****************************************************************\n");

  GenVCQSVector genvcqs = cir->getGenVCQSvec();

  // Set row/column numbers in terminals and get number of Transient-used
  // terminals.
  int nterms = cir->setMatrixRCs();
  // Get rows numbers of desired outputs
  om->getRowNumbers(cir, iv);

  // ***********************  Create matrix and vectors
  // Ground node not included in nrcs
  int nrcs = nterms - 1;
  // Number of nonlinear charges in circuit
  int nCharges = genvcqs.size();
  // Nonlinear charges vector
  double *q = new double[nCharges + 1];
  // NAM matrix. No need to clean matrix because it is fresh.
  UMFPackDI nam(nrcs, 24*nrcs);
  // Source vector (with ground)
  double *svec_ground = new double[nrcs+1];
  // leave extra ground element out (svec[-1])
  double *svec = svec_ground + 1;
  // unknowns vector (with ground)
  double *uvec_ground = new double[nrcs+1];
  // Ground node potential is zero
  uvec_ground[0] = 0.;
  // This would be uvec_(j+1)
  double *uvec = uvec_ground + 1;

  //*********************************MY DEFINED VARIABLES*****************
  int &im = iparm[IM];
  int &dcop = iparm[DCOP];
  int &vartstep = iparm[VARTSTEP];

  int iter(0);
  double residual(0.);

  printf("\nMatrix size: %d\n", nrcs);
  printf("\n Time \t\t Time Steps \t Average Iter. \t Max. Residual\n");
  printf("---------------------------------------------------------------\n");

  if (dcop) {
    // Calculate DC operating point
    int rejections;
    OP opo( cir );     			// Operating Point object   
    residual = opo.op(nrcs, nam, uvec, svec, iter, rejections, q);

    if(residual >= 1. ) {
      printf( "DC Operating Point could not converge. Residual > 1.");
      printf( "Abrupt end to the program. Residual = %.5E \n", residual );
      exit(EXIT_FAILURE);  
    }
    nam.reset();
  } else {
    // Set initial conditions
    for (int i(0); i<nrcs; i++){
      uvec[i] = 0.;
    }
    for( int i(0); i <= nCharges; i++ ){
      q[i] = 0.;
    }
  }

  if( im == 1) {
    // To initialise the Trapezoidal i.m. we need to obtain the
    // derivatives first.
    // Create Integration Method Object (imo)
    BEuler imo_be(nrcs, nCharges);
    imo_be.advance(uvec,q);
    TimeMarching<BEuler> tm_be(cir, nrcs, &imo_be, nam);

    double *uvec1_ground = new double[nrcs+1];
    double *uvec1 = uvec1_ground + 1;
    double *q1 = new double[nCharges];

    for (int i(-1); i < nrcs; i++)
      uvec1[i] = uvec[i];
    for (int i(0); i < nCharges; i++)
      q1[i] = q[i];

    double tstep1 = .1 * dparm[TSTEP];

    // Flag to indicate if the time step has changed
    bool newtstep(true);
    tm_be.step(nam, uvec1, q1, svec, 0., tstep1, newtstep, residual, iter);

    // Reset NAM
    nam.reset();

    // Calculate derivatives manually in uvec1 and q1
    for (int i(-1); i < nrcs; i++)
      uvec1[i] = (uvec1[i] - uvec[i]) / tstep1;
    for (int i(0); i < nCharges; i++)
      q1[i] = (q1[i] - q[i]) / tstep1;
    
    Trapezoidal imo(nrcs, nCharges);
    // Pass initial conditions to integration method.
    imo.init(uvec, q, uvec1, q1, dparm[TSTEP]);
    // These are no longer needed
    delete [] uvec1_ground;
    delete [] q1;

    if ( vartstep == 1)
    	variableTimeStep(cir, nrcs, &imo, nam, uvec, q, svec);
    else
    	fixedTimeStep(cir, nrcs, &imo, nam, uvec, q, svec);

  } else {
    // Create Integration Method Object (imo)
    BEuler imo(nrcs, nCharges);
    imo.init(uvec, q, dparm[TSTEP]);

    if ( vartstep == 1)
	variableTimeStep(cir, nrcs, &imo, nam, uvec, q, svec);
    else
	fixedTimeStep(cir, nrcs, &imo, nam, uvec, q, svec);

  } // END OF IF-STATEMENT

  printf( "Total number of time steps = %d \n", counter );
  printf("\n");

  printf("****************** End Of Transient Analysis *******************\n");

  delete [] q;
  delete [] uvec_ground;
  delete [] svec_ground;

}

