// This may look like C code, but it is really -*- C++ -*-
//
// Regular Transient analysis. 
// 
// by Jahanzeb Tariq Khan and Carlos Christoffersen

#ifndef Transient_H
#define Transient_H

#include <string>
#include <cstdio>
#include <carrot/Analysis.h>
#include <carrot/UMFPackDI.h>
#include "TimeMarching.h"

class Circuit;

class Transient : public Analysis
{
public:

  // ParmSet takes care of parameter handling
  Transient(ParmSet* pset1) : Analysis(pset1) {}

  virtual ~Transient() {}

  // The main method to actually run the analysis
  // (Decide what to do with the results)
  virtual void run(Circuit* cir, OutputManager* om);

private:

  // Save the contents of one time sample
  void saveVector(const double& t, double* uvec, 
		  const int& iter, const double& residual);

  template <typename IntegMethod>
  void fixedTimeStep(Circuit* cir, const int& nrcs, 
		     IntegMethod* imo, UMFPackDI& nam, 
		     double* uvec, double* q, double* svec)
  {
    double *dparm = getParmSet()->getDHandler()->getParmVec();
    double &tstop = dparm[TSTOP];
    double &tstep = dparm[TSTEP];

    TimeMarching<IntegMethod> tm(cir, nrcs, imo, nam);
    // *********************************************************************
    // Main Loop follows
    // *********************************************************************
    bool newtstep = true;
    for(double t = 0; t < tstop; t += tstep )
    {
      double residual = 1.;
      int iter;

      tm.step(nam, uvec, q, svec, t, tstep, newtstep, residual, iter);

      // Tell the integration object that a time step is finished.
      imo->advance(uvec, q);

      // Save voltages and print status
      double tsave = t+tstep;
      saveVector(tsave, uvec, iter, residual);

      // Time step never changes (as the name of this method suggest)
      newtstep = false;

    } // END OF FOR LOOP - TIME
  }  

  template <typename IntegMethod>
  void variableTimeStep(Circuit* cir, const int& nrcs,
                     IntegMethod* imo, UMFPackDI& nam,
                     double* uvec, double* q, double* svec)
  {
    double *dparm = getParmSet()->getDHandler()->getParmVec();
    double &tstop = dparm[TSTOP];
    double &tstep_max = dparm[TSTEP];

    TimeMarching<IntegMethod> tm(cir, nrcs, imo, nam);
    bool newtstep = true;
    double t = 0.;
    double tstep = tstep_max;
    double &tstep_min = dparm[MINTSTEP];
    double &maxdelta = dparm[MAXDELTA];
    int rejections = 0;
    double *usave = new double[nrcs];

    // *********************************************************************
    // Main Loop follows
    // *********************************************************************
    do {
      double residual = 1.;
      int iter;
      double delta;
      double tstep_new;
      
      // Save node voltages
      for (int j=0; j<nrcs; j++)
	usave[j] = uvec[j];

      bool rc = 
	tm.step(nam, uvec, q, svec, t, tstep, newtstep, residual, iter);
      if( rc )
	imo->getTruncError(uvec, delta, tstep_new, maxdelta);
      else 
	// Set delta to a big value to reduce time step
	delta = 10.;

      if (delta > maxdelta) {
	// Something went wrong
	if (tstep > tstep_min ) {
	  // We can still reduce the time step
	  // Time step rejected
	  tstep = .5 * tstep;
	  rejections += 1;
	  newtstep = true;
	  // Restore node voltages
	  for (int j=0; j<nrcs; j++)
	    uvec[j] = usave[j];

	} else {
	  // We have to give up
	  if (!rc) {
	    printf( "No convergence. Residual = %e \n", residual );
	    exit(-1);
	  }
	  else {
	    printf( "Warning: truncation error = %g \n", delta);
	    // Time step accepted
	    t += tstep;
	    // Tell the integration object that a time step is finished.
	    imo->advance(uvec, q);
	    // Save voltages and print status
	    saveVector(t, uvec, iter, residual);
	    newtstep = false;
	  }
	}
      } else {
	// Time step accepted
	t += tstep;
	// Tell the integration object that a time step is finished.
	imo->advance(uvec, q);
	// Save voltages and print status
	saveVector(t, uvec, iter, residual);
	// Prepare next time step
	double tstep_o = tstep;
	if (tstep_new > tstep_min)
	  if (tstep_new < tstep_max)
	    tstep = tstep_new;
	  else
	    tstep = tstep_max;
	else
	  tstep = tstep_min;
	// Set flag
	if (tstep != tstep_o)
	  newtstep = true;
      }
    } while ( t <= tstop );  // END OF LOOP - TIME

    printf( "Total number of rejections = %d \n", rejections );
    delete [] usave;
  }


  // Variables for output 
  OutputManager* om;
  std::vector<int> iv;
  int counter;
  int iter_acc;
  double res_norm;

};

#endif
