// This may look like C code, but it is really -*- C++ -*-
// Implements Trapezoidal integration method
// by Jahanzeb Tariq Khan

#ifndef Trapezoidal_H
#define Trapezoidal_H

class Trapezoidal
{
public:

  Trapezoidal(const int& nrcs, const int& ncharges)
    : nrcs(nrcs), ncharges(ncharges), factor(0.), 
        tstep(0.), t1(0.), t2(0.), t3(0.)
  {
    double *uvecn_ground = new double[nrcs + 1];
    uvecn_ground[0] = 0.;
    uvecn = uvecn_ground + 1;
    double *dvn_ground = new double[nrcs + 1];
    dvn_ground[0] = 0.;
    dvn = dvn_ground + 1;
    qn = new double[ncharges];
    dqn = new double[ncharges];
    uvecn1 = new double[nrcs];
  }

  ~Trapezoidal()
  {
    delete [] uvecn1;
    delete [] dqn;
    delete [] qn;
    delete [] (dvn-1);
    delete [] (uvecn-1);
  }

  // Set time step value
  inline void seth(const double& h)
  {
    tstep = h;
    factor = 2. / h;
  }

  // Returns the factor to multiply the C matrix.
  inline double getCfactor()
  {
    return factor;
  }

  // Returns the current contribution of a linear capacitance
  inline double getCieq(const int& col1, const int& col2, const double& c)
  {
    return factor * c * (uvecn[col1] - uvecn[col2]) 
      + c * (dvn[col1] - dvn[col2]); 
  }

  // Returns charge contribution only
  inline void getIeqQ(const double& qj, const int& i,
		      double& ieq)
  {
    ieq = (qn[i] - qj) * factor + dqn[i];
  }

  // Returns ieq and geq for each controlling voltage
  inline void getIeqGeq(const double& dqj,
			const double& vj, 
			double& ieq, double& geq)
  {
    ieq = (dqj * vj) * factor;
    geq = dqj * factor;
  }

  // Save previous values for next time step
  void advance(const double* uvec, const double* q)
  {
    t1 = t2;
    t2 = t3;
    t3 += tstep;

    for (int i=0; i<nrcs; i++) {
      dvn[i] *= -1.;
      dvn[i] += factor * (uvec[i] - uvecn[i]);
    }
    for( int i=0; i<ncharges; i++) {
      dqn[i] *= -1.;
      dqn[i] += factor * (q[i] - qn[i]);
    }

    for (int i=0; i<nrcs; i++){
      uvecn1[i] = uvecn[i];
      uvecn[i] = uvec[i];
    }
    for (int i=0; i<ncharges; i++)
      qn[i] = q[i];
  }

  // Initialise derivatives 
  void init(const double* uvec, const double* q,
	    const double* dv, const double* dq, const double& h)
  {
    t2 = tstep = h;
    for (int i=0; i<nrcs; i++) {
      uvecn1[i] =  uvecn[i] = uvec[i];
      dvn[i] = dv[i];
    }
    for (int i=0; i<ncharges; i++) {
      qn[i] = q[i];
      dqn[i] = dq[i];
    }
  }

  inline void getTruncError(const double* uvec, 
			    double& delta, double& newtstep, double& maxdelta)
  {
    delta = 0.;
    for( int i(0); i < nrcs; i++ ) {
      double truncError;
      // Only check if variable has a meningful value
      if ((fabs(uvec[i]) + fabs(uvecn[i]) + fabs(uvecn1[i])) > 1e-10) {
	// Do the extrapolation, etc.
	double a = - ( uvecn[i] - uvecn1[i] - (dvn[i])*(t2 - t1) ) / 
	  ( t2*t2 + t1*t1 - 2*t1*t2);
	double b = dvn[i] - 2*a*t2;
	double c = uvecn[i] - a*t2*t2 - b*t2;
	
	truncError = fabs(uvec[i] - (a * t3*t3 + b * t3 + c) );
	
	// Make a relative error (if possible)
	double v = fabs(uvec[i]);
	if (v > 1.)
	  truncError /= v;
      }
      if ( truncError > delta )
	delta = truncError;
	  
    }
    newtstep = tstep * sqrt(maxdelta / delta);
  }


private:

  int nrcs;
  int ncharges;
  double factor;
  double tstep;
  double t1, t2, t3;

  double* dqn;
  double* dvn;

  double* uvecn;
  double* uvecn1;
  double* qn;

};

#endif
