#include <cassert>
#include <vector>
#include <carrot/UMFPackDI.h>

extern "C" {
#include <suitesparse/umfpack.h>
}

#include <cstdio>

UMFPackDI::~UMFPackDI()
{
  if (factor_flag)
    umfpack_di_free_numeric(&Numeric);
  if (symbolic_flag) 
    umfpack_di_free_symbolic(&Symbolic);
  if (flag) {
    delete [] W;
    delete [] Wi;

    delete [] colPtr;
    delete [] rowIdx;
    delete [] valA;
    delete [] Map;
  } 
}


void UMFPackDI::buildCompressed()
{
  int nz = rowvec.size();
  if (!flag) {
    // Allocate space for compressed col. format, generate matrix and
    // generate Map
    colPtr = new int[nrcs+1];
    rowIdx = new int[nz];
    valA = new double[nz];
    Map = new int[nz];
    int status = umfpack_di_triplet_to_col(nrcs, nrcs, nz, 
					   &(rowvec[0]), 
					   &(colvec[0]), &(valuevec[0]),
					   colPtr, rowIdx, valA, Map) ;
    assert(status == UMFPACK_OK);
    // Allocate space for workspace vectors
    Wi = new int[nrcs];
    W = new double[5 * nrcs];
    flag = true;
  }
  else {
    // colPtr and rowIdx should be the same. Recalculate valA from Map.
    for (int p(0) ; p < colPtr[nrcs] ; p++) 
      valA[p] = 0.;
    for (int k(0) ; k < nz ; k++) 
      valA[Map[k]] += valuevec[k] ;
  }
}

int UMFPackDI::symbolic()
{
  assert(flag);
  if (!symbolic_flag) {
    int status = umfpack_di_symbolic(nrcs, nrcs, colPtr, rowIdx, valA, 
				     &Symbolic, NULL, NULL);
    //  assert(status == UMFPACK_OK);
    symbolic_flag = true;
    return status;
  } else 
    return 0;
}

int UMFPackDI::factor()
{
  assert(flag);
  int status = umfpack_di_numeric(colPtr, rowIdx, valA,
				  Symbolic, &Numeric, NULL, NULL);
  // assert(status == UMFPACK_OK);
  factor_flag = true;
  return status;
}

int UMFPackDI::solve(double* x, double* rhs)
{
  assert(flag);
  int status = umfpack_di_wsolve(UMFPACK_A, 
				colPtr, rowIdx, valA,
				x, rhs, Numeric, NULL, NULL, Wi, W);
  // assert(status == UMFPACK_OK);
  return status;
}

int UMFPackDI::solveTransposed(double* x, double* rhs)
{
  assert(flag);
  double* Control = NULL;
//   double Control[UMFPACK_CONTROL];
//   umfpack_di_defaults(Control);
//   Control[UMFPACK_PRL] = 3;
//   Control[UMFPACK_PIVOT_TOLERANCE] = 1e-3;
//   Control[UMFPACK_IRSTEP] = 10;
//   Control[UMFPACK_AGGRESSIVE] = 0;
  int status = umfpack_di_wsolve(UMFPACK_At, 
 				 colPtr, rowIdx, valA,
 				 x, rhs, Numeric, Control, NULL, Wi, W);
//   umfpack_di_report_status(Control, status);

//   // Test transposing the matrix manually
//   printf("\n\nAhoy!\n\n");
//   int nz = rowvec.size();
//   int* Rp = new int[nrcs+1];
//   int* Ri = new int[nz];
//   double* Rx = new double[nz];
//   int status = umfpack_di_transpose (nrcs, nrcs, colPtr, rowIdx, valA, 
// 				     NULL, NULL, Rp, Ri, Rx);
//   umfpack_di_report_status(Control, status);

//   status = umfpack_di_symbolic(nrcs, nrcs, Rp, Ri, Rx, 
//  				   &Symbolic, NULL, NULL);
//   umfpack_di_report_status(Control, status);

//   status = umfpack_di_numeric(Rp, Ri, Rx,
//  			      Symbolic, &Numeric, NULL, NULL);
//   umfpack_di_report_status(Control, status);

//   status = umfpack_di_wsolve(UMFPACK_A, 
//  			     Rp, Ri, Rx,
// 			     x, rhs, Numeric, NULL, NULL, Wi, W);
//   umfpack_di_report_status(Control, status);
   assert(status == UMFPACK_OK);

//   // Create a dense matrix
//   int nc2 = nrcs * nrcs;
//   double *A = new double[nc2];
//   for (int i(0); i<nc2; i++)
//     A[i] = 0.;
//   for (unsigned i(0); i < rowvec.size(); i++)
//     A[rowvec[i] + colvec[i] * nrcs] += valuevec[i];
//   // Print transposed multiplication 
//   printf("\n");
//   for (int i(0); i<nrcs; i++) {
//     double sum = 0.;
//     for (int j(0); j<nrcs; j++) 
//       sum += A[nrcs * i + j] * x[j];
//     printf("[%d] %g \n", i, sum);
//   }

  return status;
}

void UMFPackDI::clearNonlinearNumeric()
{
  if (factor_flag) {
    umfpack_di_free_numeric(&Numeric);
    // Erase only the nonlinear part of the triplet structure
    rowvec.resize(triplet_lin_size);
    colvec.resize(triplet_lin_size);
    valuevec.resize(triplet_lin_size);
    factor_flag = false;
  }
}

void UMFPackDI::clearCapNumeric()
{
  if (factor_flag) {
    umfpack_di_free_numeric(&Numeric);
    // Erase the nonlinear part and capacitors of the triplet structure
    rowvec.resize(triplet_g_size);
    colvec.resize(triplet_g_size);
    valuevec.resize(triplet_g_size);
    factor_flag = false;
  }
}

void UMFPackDI::clearCompressed()
{
  if (flag) {
    delete [] colPtr;
    delete [] rowIdx;
    delete [] valA;
    delete [] Map;
    flag = false;
  } 
  if (symbolic_flag) { 
    umfpack_di_free_symbolic(&Symbolic);
    symbolic_flag = false;
  }
}

void UMFPackDI::reset()
{
  // Erase the nonlinear part and capacitors of the triplet structure
  rowvec.resize(0);
  colvec.resize(0);
  valuevec.resize(0);
  if (factor_flag) {
    umfpack_di_free_numeric(&Numeric);
    factor_flag = false;
  }
  if (flag) {
    delete [] colPtr;
    delete [] rowIdx;
    delete [] valA;
    delete [] Map;
    flag = false;
  } 
  if (symbolic_flag) { 
    umfpack_di_free_symbolic(&Symbolic);
    symbolic_flag = false;
  }
}


void UMFPackDI::printTriplet()
{
  // This is useful for small matrices only
  if (nrcs < 31) {
    // Create a dense matrix
    int nc2 = nrcs * nrcs;
    double *A = new double[nc2];
    for (int i(0); i<nc2; i++)
      A[i] = 0.;
    for (unsigned i(0); i < rowvec.size(); i++)
      A[rowvec[i] + colvec[i] * nrcs] += valuevec[i];
    
    printf("\n[");
    for (int i(0); i<nrcs; i++) {
      for (int j(0); j<nrcs; j++)
	if (A[i + nrcs*j] != 0.)
	  printf("%g ", A[i + nrcs*j]);
	else
	  printf("0 ");
      printf(";\n");
    }
    printf("]\n");
  }
  else {
    printf("Matrix too large (%d), not printing\n", nrcs);
  }
}
