// This may look like C code, but it is really -*- C++ -*-
//
// Common variables used in BJT model
//


#ifndef CommonParams_H
#define CommonParams_H 1

// Intermediate results needed in several current sources

// Used for Ibf, Ibr, Ile, Ilc
template<typename outDouble, typename voltDouble, 
	 typename paramDouble, typename varDouble>
inline void getIexp(outDouble& iexp, voltDouble& v, 
		    varDouble& is, paramDouble& n, varDouble& vt)
{
  iexp = is * (safe_exp(v / n / vt) - 1.);
}

template<typename outDouble, typename voltDouble, 
	 typename paramDouble, typename varDouble>
void getKqb(outDouble& kqb, voltDouble& vbe, voltDouble& vbc, 
	    outDouble& ibf, outDouble& ibr,
	    paramDouble* dparm, varDouble& is, 
	    varDouble& vt)
{
  // q1m1 = 1/q1
  outDouble q1m1 = 1.;
  if (dparm[VAR] != 0.)
    q1m1 -= vbe / dparm[VAR];
  if (dparm[VAF] != 0.)
    q1m1 -= vbc / dparm[VAF];
  kqb = 1. / q1m1;
  if (dparm[IKF] != 0.) {
    outDouble q2 = ibf / dparm[IKF] + ibr / dparm[IKF];
    kqb *= .5 * (1. + sqrt(1. + 4. * q2));
  }
}

template<typename outDouble, typename tempDouble>
inline void getVt(outDouble& vt, tempDouble& T)
{
  vt = kBoltzman * T / eCharge;
}


// Used for IS, ISE, ISC and ISS
template<typename outDouble, typename paramDouble, 
	 typename varDouble, typename tempDouble>
void getIs(outDouble& is,
	   paramDouble& IS, paramDouble& N, varDouble& vt, 
	   paramDouble* dparm, tempDouble& T)
{
  outDouble tn = T / (dparm[TNOM]+tzero);
  is = IS * exp((tn-1.) * dparm[EG] / vt) * pow(tn, dparm[XTI]/N);
}

template<typename outDouble, typename paramDouble, typename tempDouble>
inline void getEgap(outDouble& egap, paramDouble* dparm, tempDouble& T)
{
  egap = dparm[EG] - .000702 * T*T / (T+1108);
}

// Used for VJE, VJC, VJS
template<typename outDouble, typename paramDouble, 
	 typename varDouble, typename cvarDouble,
	 typename tempDouble>
void getVbi(outDouble& Vbi, varDouble& vt, cvarDouble& egapnom,
	    paramDouble& VJ, paramDouble* dparm, 
	    tempDouble& T)
{
  outDouble tn = T / (dparm[TNOM]+tzero);
  outDouble egap;
  getEgap(egap, dparm, T);
  Vbi = VJ*tn - 3.*vt*log(tn) - tn * egapnom + egap;
}

// Used for CJE, CJC, CJS
template<typename outDouble, typename paramDouble, 
	 typename varDouble, typename tempDouble>
void getCbi(outDouble& Cbi, varDouble& vt,
	    varDouble& Vbi, paramDouble& VJ,
	    paramDouble& CJ, paramDouble& MJ, 
	    paramDouble* dparm, tempDouble& T)
{
  Cbi = CJ * (1. + MJ * (.0004 * (T-(dparm[TNOM]+tzero)) + 1. - Vbi / VJ));
}

// Used for BF, BR
template<typename outDouble, typename paramDouble, typename tempDouble>
inline void getB(outDouble& bt, paramDouble& B, 
		 paramDouble* dparm, tempDouble& T)
{
  outDouble tn = T / (dparm[TNOM]+tzero);
  bt = B * pow(tn, dparm[XTB]);
}

// Used for Qbej, Qbcj, Qbxj (depletion charge only, must add diffusion)
template<typename outDouble, typename voltDouble, typename paramDouble, 
	 typename varDouble, typename cvarDouble>
void getQj(outDouble& qj, voltDouble& v, varDouble& VJ,
	   varDouble& CJ, varDouble& k5, varDouble& k6, varDouble& k7, 
	   paramDouble& MJ, paramDouble& FC,  cvarDouble& k4)
{
  if (CJ > 0.) {
    if (v < FC * VJ) {
      qj = k5 * (1. - pow(1. - v / VJ, k4));
    } else {
      qj = k6 * (((1. - FC*(1. + MJ)) * v + .5*MJ*v*v/VJ) - k7)
	+ k5 * (1. - pow(1.-FC, k4));
    }
  } else 
    qj = 0.;
}

// Used for all depletion capacitances
template<typename outDouble, typename paramDouble, 
	 typename varDouble, typename cvarDouble>
void getk567(outDouble& k5, outDouble& k6, outDouble& k7,
	     varDouble& VJ, varDouble& CJ, 
	     paramDouble& MJ, paramDouble& FC,  cvarDouble& k4)
{
  k5 = VJ * CJ / k4;
  k6 = CJ * pow(1. - FC, - MJ - 1.);
  k7 = ((1. - FC * (1.+MJ)) * VJ * FC + .5*MJ * VJ*FC*FC);
}


#endif
