// This may look like C code, but it is really -*- C++ -*-
//
// Implements the charge stored in the BE junction. Needs VBE and VBC
// to calculate effect of currents in TF.
//
//             
//          o  1              
//          |                            
//        --+--
//               QBE  
//        --+--
//          |                          
//          o  0    	      
//                
// 

#ifndef QBE_H
#define QBE_H 1

class QBE_eval {
public:
  // This are locally-defined variables
  enum {T_IS, VT, T_VJE, T_CJE, k5, k6, k7, MAX_TVAR_INDEX};
  enum {k4, EGAPn, MAX_CVAR_INDEX};

  QBE_eval(int *iparm) {} // iparm not needed here

  // Main evaluation function. Only voltage dependence considered here.
  template <typename currDouble, typename voltDouble, typename paramDouble, 
	    typename varDouble, typename cvarDouble>
  inline void operator()(currDouble& qbe, voltDouble* v, 
			 paramDouble* dparm, 
			 varDouble* tvar, cvarDouble* cvar)
  {
    // Depletion charge
    getQj(qbe, v[0], tvar[T_VJE], tvar[T_CJE], 
	  tvar[k5], tvar[k6], tvar[k7], dparm[MJE], dparm[FC], cvar[k4]);
    if (dparm[TF] > 0.) {
      // Need to calculate ibf for TT component
      currDouble ibf;
      getIexp(ibf, v[0], tvar[T_IS], dparm[NF], tvar[VT]);
      currDouble ibr;
      getIexp(ibr, v[1], tvar[T_IS], dparm[NR], tvar[VT]);
      currDouble kqb;
      getKqb(kqb, v[0], v[1], ibf, ibr, dparm, tvar[T_IS], tvar[VT]);
      // Calculate effective tf
      currDouble tf = dparm[TF];
      if (dparm[VTF] > 0.) {
	// Calculate x
	currDouble x = ibf / (ibf + dparm[ITF]);
	tf *= (1. + dparm[XTF] * x*x 
	  * safe_exp(v[1] /1.44 /dparm[VTF]));
      }
      qbe += tf * ibf / kqb;
    }
    // Area effect
    qbe *= dparm[AREA];
  }

  // Adjust variables to a new temperature.
  template <typename tvarDouble, typename varDouble, typename tempDouble>
  inline void setTemp(tvarDouble* tvar, 
		      varDouble* dparm, varDouble* cvar, tempDouble& T)
  {
    tempDouble Tabs = T + tzero;
    // Parameter calculation 
    getVt(tvar[VT], Tabs);
    getIs(tvar[T_IS], dparm[IS], dparm[NF], tvar[VT], dparm, Tabs);
    getVbi(tvar[T_VJE], tvar[VT], cvar[EGAPn], dparm[VJE], dparm, Tabs);
    getCbi(tvar[T_CJE], tvar[VT], 
	   tvar[T_VJE], dparm[VJE],
	   dparm[CJE], dparm[MJE], dparm, Tabs);
    getk567(tvar[k5], tvar[k6], tvar[k7], 
	    tvar[T_VJE], tvar[T_CJE],
	    dparm[MJE], dparm[FC], cvar[k4]);
  }
  
  // For efficiency, we may want to pre-calculate some expressions
  // that do not change with voltage nor temperature.  Adjust variables
  // to a new parameer set.
  template <typename varDouble>
  inline void newParms(varDouble* cvar, varDouble* dparm)
  {
    cvar[k4] = 1. - dparm[MJE];
    varDouble tnom = dparm[TNOM]+tzero;
    getEgap(cvar[EGAPn], dparm, tnom);
  }

  //private:
  
  // Pointer to integer parameter vector from macroelement
  //  int *iparm; // not necessary for this model

};


typedef FADBADGenericVCS<QBE_eval, GenericVCQS, 2, 
			 MAX_DPARM_INDEX, TEMP, false> QBE;
// With thermal port
typedef FADBADTGenericVCS<QBE_eval, GenericVCQS, 3, 
			  MAX_DPARM_INDEX, TEMP> TQBE;

#endif
