#include "global.h"
#include "MOSFETL3.h"
#include <carrot/Options.h>
#include <carrot/FADBADGenericVCS.h> 
#include <carrot/FADBADTGenericVCS.h> 
#include <carrot/FADBADLinearVCS.h> 
#include <carrot/FADBADPowerSource.h> 

static ItemInfo MOSFETL3_info = {
  "mosfetl3",
  "MOSFET Level 3 model",
  "Justin Chen"
  "no"
};

// Double Parameters
// Make sure enum and ParInfo definitions are in the same order!
enum {L, W, AD, AS, PD, PS, IS, JS, PBSW, 
      CJ, CJSW, MJ, MJSW, FC, PB, CGSO, CGDO, CGBO, LD, WD,
      LAMB, TOX, EOX, NI, ES, NSUBN, NSUBP, U0N, U0P, TNOM,
      REFTEMP, KAPA, TEMP, NSS, XJ, ETA, DELT, 
      NFS, VMAX, THET, DELTA, VT0P, VT0N, VTC, MAX_DPARM_INDEX};

//Insert ParmInfo values here
static ParamHandler<double>::ParmInfo MOSFETL3_dparams[] = { 
{"l", "Channel length (m)", false, 100e-6},
{"w", "CHannel width (m)", false, 200e-6},
{"ad", "Drain diffusion area (m^2)", false, 0.},
{"as", "Source diffustion area (m^2)", false, 0.},
{"pd", "Drain diffusion parameter", false, 0.},
{"ps", "Source diffusion parameter", false, 0.},
{"is", "Bulk-junction saturation current (amp)", false, 1e-14},
{"js", "Bulk-junction saturation current per area (amp/m^2)", false, 0.},
{"pbsw", "Bulk-junction sidewall potential (V)", false, 0.8},
{"cj", "Bulk-junction zero-bias bottom capacitance per area (Farad/m^2)", false, 0},
{"cjsw", "Bulk-junction zero-bias sidewall capacitance per length (Farad/m)", false, 0.},
{"mj", "Bulk-junction bottom grading coefficient", false, 0.5},
{"mjsw", "Bulk-junction sidewall grading coefficient", false, 0.33},
{"fc", "Bulk-junction forward-bias capacitance coefficient", false, 0.5},
{"pb", "Bulk-junction bottom potential", false, 0.8},
{"cgso", "Gate-source overlap capacitance per channel width (Farad/m)", false, 0.},
{"cgdo", "Gate-drain overlap capacitance per channel width (Farad/m)", false, 0.},
{"cgbo", "Gate-bulk overlap capacitance per channel length (Farad/m)", false, 0.},
{"ld", "Lateral diffusion length (m)", false, 0.},
{"wd", "Lateral diffusion width (m)", false, 0.},
{"lamb", "Channel-length modulation (1/volt)", false, 0.},
{"tox", "Thin oxide thickness (m)", false, 1e-7},
{"eox", "Permittivity of oxide (Farad/m)", false, 3.453e-11},
{"ni", "Intrinsic carrier concentration for Si (m^-3)", false, 1.45e16},
{"es", "Permittivity of Si (Farad/m)", false, 1.0359e-10},
{"nsubn", "Surstrate doping density (1/m^3)", false, 0.},
{"nsubp", "Surstrate doping density (1/m^3)", false, 0.},
{"u0n", "NMOS Surface mobility (m^2/(V*sec))", false, 0.06},
{"u0p", "PMOS Surface mobility (m^2/(V*sec))", false, 0.06},
{"tnom", "Nominal temperature (C)", false, 27.},
{"reftemp", "Reference temperature (C)", false, 27.},
{"kapa", "Saturation field factor", false, 0.2},
{"temp", "Temperature (C). Defaults to global temp.", false, 27.},
{"nss", "Surface state density (1/cm^2)", false, 0.},
{"xj", "source, drain junction depth (um)", false, 0.},
{"eta", "Static feedback on threshold voltage", false, 0.},
{"delt", "Width effect on threshold voltage", false, 0.},
{"nfs", "Fast surface state density (1/m^2)", false, 0.},
{"vmax", "Maximum drift velocity of carriers", false, 0.},
{"thet", "Mobility modulation", false, 0.},
{"delta", "Width effect on threshold voltage", false, 0.},
{"vt0p", "PMOS zero bias threshold voltage", false, 0.},
{"vt0n", "NMOS zero bias threshold voltage", false, 0.},
{"vtc", "Threshold voltage choice", false, 0.}

};

// Uncomment for integer parameters
//enum {MAX_IPARM_INDEX};
//static ParamHandler<int>::ParmInfo MOSFETL3_iparams[] = {  };


// Integer parameters
// ***********************************************************************
enum {TYPE, MAX_IPARM_INDEX};
static ParamHandler<int>::ParmInfo MOSFETL3_iparams[] = {
  {"type", "0 (nmosfetl3) or 1 (pmosfetl3)", false, 0},
  
};



// Include current source models

#include "IbdL3.h"
#include "IbsL3.h"
#include "Ids1L3.h"
#include "IdbL3.h"
#include "IsbL3.h"
#include "Isd1L3.h"

#include "Qdn.h"
#include "Qsn.h"
#include "Qbn.h"
#include "Qbsn.h"
#include "Qbdn.h"

#include "Qdp.h"
#include "Qsp.h"
#include "Qbp.h"
#include "Qsbp.h"
#include "Qdbp.h"



//Library loading mechanism
extern "C" {
  const char * ModuleName(){
    return MOSFETL3_info.type.c_str();
  }
  ParmSet * newParmSet(const char * model){
    return new ParmSet(model,
		         &MOSFETL3_info,
		         MOSFETL3_dparams, MAX_DPARM_INDEX, 
		         MOSFETL3_iparams, MAX_IPARM_INDEX);
  }
  void delParmSet(ParmSet * p){
    delete p;
    p=0;
  }
  MacroElement * newObject(ParmSet * p, const char * name){
    return new MOSFETL3(p, name);
  }
  void delObject(MacroElement * e){
    delete e;
    e = 0;
  }
};

MOSFETL3::MOSFETL3(ParmSet *pset, const std::string& name)
  :MacroElement(pset, name, 4)
{
  //Constructor code goes here (above, 4 is the number of terminals)
  //printf("-----%d------\n\n", MAX_DPARM_INDEX);
}

MOSFETL3::~MOSFETL3()
{
  //Destructor code goes here
}

void MOSFETL3::init()
{
  // access to double parameters
  double *dparm = getParmSet()->getDHandler()->getParmVec();
  int *iparm = getParmSet()->getIHandler()->getParmVec();
  
  // Check required parameters
  getParmSet()->checkParms();
  // Check external terminal connections
  assert(checkTerms());
  // Set number of internal nodes
  setNInternalNodes(0);
  // 1. Create individual CSs
  // LinearVCCS* lvccs = new LinearVCCS(dparm[P1]);
  // Insert into MacroElement internal lists
  // insert(lvccs);
  // 2. connect internal terminals.
  // Vector with internal terminal connections (must be static)
  // In the example output and controlling terminals are
  // connected to external terminals 0 and 1.
  // static const int tvec[4] = {0,1,0,1};
  // lvccs->setTermNum(tvec);


// Set default temperature if not specified in parameter set
  if (!(getParmSet()->getDHandler()->isSet(TEMP)))
    getParmSet()->getDHandler()->
      setParm(TEMP, options.getDHandler()->getParmVec()[OPT_TEMP]);
  
 // Assign connection vector
  if (iparm[TYPE] == 0) {
  
  static const int ibdL3terms[4] = {3,0,3,0};
  static const int ibsL3terms[4] = {3,2,3,2};
  static const int ids1L3terms[8] = {0,2,1,2,0,2,3,2};
 

  GenericVCCS* ibdL3 = new IbdL3(dparm, NULL);
  GenericVCCS* ibsL3 = new IbsL3(dparm, NULL);
  GenericVCCS* ids1L3 = new Ids1L3(dparm, NULL);
  
  insert(ibdL3, ibdL3terms);
  insert(ibsL3, ibsL3terms);
  insert(ids1L3, ids1L3terms);
  
  static const int qdnterms[8] = {0,1,1,2,0,2,3,2};
  static const int qsnterms[8] = {2,1,1,2,0,2,3,2};
  static const int qbnterms[8] = {3,1,1,2,0,2,3,2};
  static const int qbsnterms[4] = {3,2,3,2};
  static const int qbdnterms[4] = {3,0,3,0};
  
  GenericVCQS* qdn = new Qdn(dparm, NULL);
  GenericVCQS* qsn = new Qsn(dparm, NULL);
  GenericVCQS* qbn = new Qbn(dparm, NULL);
  GenericVCQS* qbsn = new Qbsn(dparm, NULL);
  GenericVCQS* qbdn = new Qbdn(dparm, NULL);
  
  insert(qdn, qdnterms);
  insert(qsn, qsnterms);
  insert(qbn, qbnterms);
  insert(qbsn, qbsnterms);
  insert(qbdn, qbdnterms);
  
  
  
  } else {
  
  static const int idbL3terms[4] = {0,3,0,3};
  static const int isbL3terms[4] = {2,3,2,3};
  static const int isd1L3terms[8] = {2,0,2,1,2,0,2,3};

  GenericVCCS* idbL3 = new IdbL3(dparm, NULL);
  GenericVCCS* isbL3 = new IsbL3(dparm, NULL);
  GenericVCCS* isd1L3 = new Isd1L3(dparm, NULL);

  insert(idbL3, idbL3terms);
  insert(isbL3, isbL3terms);
  insert(isd1L3, isd1L3terms);
  
  static const int qdpterms[8] = {0,1,2,1,2,0,2,3};
  static const int qspterms[8] = {2,1,2,1,2,0,2,3};
  static const int qbpterms[8] = {3,1,2,1,2,0,2,3};
  static const int qsbpterms[4] = {2,3,2,3};
  static const int qdbpterms[4] = {0,3,0,3};
  
  GenericVCQS* qdp = new Qdp(dparm, NULL);
  GenericVCQS* qsp = new Qsp(dparm, NULL);
  GenericVCQS* qbp = new Qbp(dparm, NULL);
  GenericVCQS* qsbp = new Qsbp(dparm, NULL);
  GenericVCQS* qdbp = new Qdbp(dparm, NULL);
  
  insert(qdp, qdpterms);
  insert(qsp, qspterms);
  insert(qbp, qbpterms);
  insert(qsbp, qsbpterms);
  insert(qdbp, qdbpterms);
  
  
  }
}
