#include "global.h"
#include "Resistor.h"
#include <carrot/Options.h>
#include <carrot/FADBADGenericVCS.h> 
#include <carrot/FADBADTGenericVCS.h> 
#include <carrot/FADBADLinearVCS.h> 
#include <carrot/FADBADPowerSource.h> 

// Debug
//#include <cstdio>

static ItemInfo Resistor_info = {
  "res",
  "Regular Resistor",
  "Carlos Christoffersen",
  DEFAULT_ADDRESS
};

// ***********************************************************************
// Double parameters. We need a way to correlate this with the
// information in Resistor_dparams
// ***********************************************************************
enum {R, RSH, L, W, NARROW, TNOM, TEMP, TC1, TC2, MAX_DPARM_INDEX}; 

static ParamHandler<double>::ParmInfo Resistor_dparams[] = {
  {"r", "Resistance (Ohms)", false, 1e3},
  {"rsh", "Sheet resistance (Ohms)", false, 0.},
  {"l", "Lenght (m)", false, 0.},
  {"w", "Width (m)", false, 0.},
  {"narrow", "Narrowing due to side etching (m)", false, 0.},
  {"tnom", "Nominal temperature (C)", false, 27.},
  {"temp", "Temperature (C). Defaults to global Temperature", 
   false, 27.},
  {"tc1", "Temperature coefficient 1 (Ohms/K)", false, 0.},
  {"tc2", "Temperature coefficient 2 (Ohms/K^2)", false, 0.}
};

enum {TPORT, MAX_IPARM_INDEX};
static ParamHandler<int>::ParmInfo Resistor_iparams[] = {
  {"tport", "Use thermal port (4 terminals)", false, 0}
};

extern "C" {

  //Export the name of the module we are using

  const char * ModuleName(){
    return Resistor_info.type.c_str();
  }

  ParmSet * newParmSet(const char * model){
    return new ParmSet(model, &Resistor_info, 
		       Resistor_dparams, MAX_DPARM_INDEX,
		       Resistor_iparams, MAX_IPARM_INDEX);
  }

  void delParmSet(ParmSet * p){
    delete p;
    p = 0;
  }

  MacroElement * newObject(ParmSet * p, const char * name){
    return new Resistor(p, name);
  }

  void delObject(MacroElement * e){
    delete e;
    e = 0;
  }

};

#include "RType.h"
#include "NLRType.h"

// ***********************************************************************
// Resistor members start:
// ***********************************************************************
Resistor::Resistor(ParmSet *pset, const std::string& name) 
  : MacroElement(pset, name)    
{
  int *iparm = getParmSet()->getIHandler()->getParmVec();
  if (iparm[TPORT])
    setNTerms(4);
  else
    setNTerms(2);

}

Resistor::~Resistor()
{
  // There is not much to delete here because the MacroElement class
  // takes care of deleting the CSs (that were inserted).
}


void Resistor::init()
{
  double *dparm = getParmSet()->getDHandler()->getParmVec();
  int *iparm = getParmSet()->getIHandler()->getParmVec();

  // Set default temperature if not specified in parameter set
  if (!getParmSet()->getDHandler()->isSet(TEMP))
    getParmSet()->getDHandler()->
      setParm(TEMP, options.getDHandler()->getParmVec()[OPT_TEMP]);

  // This flow should be similar for many MacroElements:
  //
  // 0. Check parameters and terminals
  // (We should really throw and exception in the following:)
  getParmSet()->checkParms();
  assert(checkTerms());
  // No internal nodes needed for now
  setNInternalNodes(0);
  // Check parameters
  if (dparm[RSH] == 0.) 
    assert(dparm[R]);
  else
    assert(dparm[W] * dparm[L]);

  if (iparm[TPORT]) {
    GenericVCCS* res = new TRType(dparm, NULL);
    static const int res_terms[6] = {0,1,0,1,2,3};
    insert(res, res_terms);
    GenericVCCS* ps = new RTypePS(dparm, NULL);
    static const int ps_terms[6] = {3,2,0,1,2,3};
    insert(ps, ps_terms);
  } else {
    LinearVCCS* lvccs = new RType(dparm, NULL);
    static const int tvec[4] = {0,1,0,1};
    insert(lvccs, tvec);
  }

}

