#include "ThermalBlock.h"
#include <carrot/Options.h>
#include "global.h"
#include <cstdlib>
#include <cstdio>

static ItemInfo ThermalBlock_info = {
  "tblock",
  "Thermal N-Port",
  "Carlos Christoffersen"
  "default"
};

// Double Parameters
// Make sure enum and ParInfo definitions are in the same order!
//enum {P1, P2, MAX_DPARM_INDEX};

//Insert ParmInfo values here
//static ParamHandler<double>::ParmInfo ThermalBlock_dparams[] = { 
//{"p1", "Parameter 1 (units)", false, 1e-3},
//{"p2", "Parameter 2 (units)", false, 0.}
//};

// Uncomment for integer parameters
enum {NPORTS, NSAMPLES, MAX_IPARM_INDEX};
static ParamHandler<int>::ParmInfo ThermalBlock_iparams[] = {  
  {"nports","Number of ports", true, 0},
  {"nsamples", "Number of samples in step response", true, 0}
};

enum {TDFILE, FDFILE, MAX_SPARM_INDEX};

static ParamHandler<std::string>::ParmInfo ThermalBlock_sparams[] = {
  {"tdfile", "Name of the file that contains the thermal step-response data", 
   true, "file1"},
  {"fdfile", "Name of the file that contains the frequency domain data", 
   true, "file2"}
};


//Library loading mechanism
extern "C" {
  const char * ModuleName(){
    return ThermalBlock_info.type.c_str();
  }
  ParmSet * newParmSet(const char * model){
    return new ParmSet(model,
		       &ThermalBlock_info,
		       NULL, 0, // No double parameters
		       ThermalBlock_iparams, MAX_IPARM_INDEX,
		       ThermalBlock_sparams, MAX_SPARM_INDEX);
  }
  void delParmSet(ParmSet * p){
    delete p;
    p=0;
  }
  MacroElement * newObject(ParmSet * p, const char * name){
    return new ThermalBlock(p, name);
  }
  void delObject(MacroElement * e){
    delete e;
    e = 0;
  }
};

#include "ThermalCS.h"
#include "ThermalVCCS.h"

ThermalBlock::ThermalBlock(ParmSet *pset, const std::string& name)
  : MacroElement(pset, name), flag(false), 
    time(0), deltaRth(0), Rth(0), frequency(0), termvec(0)
{
  // Check required parameters
  pset->checkParms();
  // access to integer parameters
  int *iparm = pset->getIHandler()->getParmVec();
  // Check external terminal connections
  assert(iparm[NPORTS] > 0);
  setNTerms(iparm[NPORTS]+1);
}

ThermalBlock::~ThermalBlock()
{
  if (flag) {
    int *iparm = getParmSet()->getIHandler()->getParmVec();
    int imax = iparm[NPORTS] * (iparm[NPORTS] + 3);
    for (int i=0; i<imax; i++) 
      delete [] termvec[i];
    delete [] termvec;
    delete [] frequency;
    delete [] Rth;
    delete [] deltaRth;
    delete [] time;
  }
}

void ThermalBlock::init()
{
  // access to integer parameters
  int *iparm = getParmSet()->getIHandler()->getParmVec();
  // String parameters
  std::string *sparm = getParmSet()->getSHandler()->getParmVec();

  assert(checkTerms());
  // Set number of internal nodes
  setNInternalNodes(iparm[NPORTS]);

  // Parse the input file
  assert(iparm[NSAMPLES] > 1);
  // Allocate space to store step response
  const int& nsamples = iparm[NSAMPLES];
  const int& nports = iparm[NPORTS];
  time = new double[nsamples];
  int blocksize = nports*(nports-1)/2 + nports;
  deltaRth = new double[(nsamples-1)*blocksize];
  double* Rth0 = new double[blocksize];

  // Open step response file 
  FILE* infile = fopen(sparm[TDFILE].c_str(), "r");
  if (infile == NULL) {
    fprintf(stderr, "Can not open thermal step-response data file: %s", 
	    sparm[TDFILE].c_str());
    exit(-1);
  }
  // Scan first block
  fscanf(infile, "%lf", time);
  for (int i=0; i<blocksize; i++) 
    fscanf(infile, "%lf", Rth0+i);
  for (int j=1; j<nsamples; j++) {
    fscanf(infile, "%lf", time+j);
    for (int i=0; i<blocksize; i++) {
      double value;
      fscanf(infile, "%lf", &value);
      // Substract previous sample
      deltaRth[nsamples-1-j+i*(nsamples-1)] = Rth0[i] - value; 
      // Save this sample
      Rth0[i] = value;
    }
  }
  fclose(infile);
  // Rth0 should now contain the first sample (DC component for
  // multitime analysis) with the format in Bill's files.

  // Open frequency-domain data file 
  infile = fopen(sparm[FDFILE].c_str(), "r");
  if (infile == NULL) {
    fprintf(stderr, "Can not open thermal frequency domain data file: %s", 
	    sparm[FDFILE].c_str());
    exit(-1);
  }
  // Scan file and count number of frequencies
  double freq;
  int nfreq(1); // Start from 1 since we also have the DC component
  while (fscanf(infile,"%lf", &freq) != EOF) {
    for (int i=0; i<blocksize; i++) {
      double dummy;
      fscanf(infile,"%lf", &dummy); // real part
      fscanf(infile,"%lf", &dummy); // imaginary part
    }
    nfreq++;
  }
  rewind(infile);
  // Allocate memory and copy DC component
  Rth = new double_complex[blocksize*nfreq];
  for (int i=0; i<blocksize; i++) {
    Rth[i*nfreq] = Rth0[i];
  }
  delete [] Rth0;
  // Now read data
  frequency = new double[nfreq];
  frequency[0] = 0.; // DC always first
  for (int j=1; j<nfreq; j++) {
    fscanf(infile,"%lf", frequency+j);
    // Data file is the step response data. Convert to impedance
    // (multiply by jw)
    for (int i=0; i<blocksize; i++) {
      double temp;
      const double_complex imag_unit(0,1);
      fscanf(infile,"%lf", &temp); // real part (step response)
      Rth[i*nfreq+j] = temp*twopi*frequency[j]*imag_unit; // imaginary part
      fscanf(infile,"%lf", &temp); // imaginary part (step response)
      Rth[i*nfreq+j] -= temp * twopi * frequency[j]; // add real part
    }
  }
  fclose(infile);
  
  // Create individual CSs Save pointers to terminal vectors (so we
  // can clean up after). Order of vectors is: first nports^2 are the
  // linear VCCS, then the nports time-delayed CS and last the
  // 2*nports sources for the gyrators.
  termvec = new int*[nports * (nports + 3)];
  int base = nports*nports;
  // Scale factor used in gyrators
  const double &scale = options.getDHandler()->getParmVec()[OPT_SCALE];
  for (int i=0; i<nports; i++) {
    // Add gyrator for this port (seems unavoidable)
    LinearVCCS* gyr1 = new LinearVCCS(scale);
    termvec[base+nports+i] = new int[4];
    termvec[base+nports+i][0] = i;
    termvec[base+nports+i][1] = nports; 
    termvec[base+nports+i][2] = nports+1+i;
    termvec[base+nports+i][3] = -1; // gnd
    insert(gyr1, termvec[base+nports+i]);
    LinearVCCS* gyr2 = new LinearVCCS(scale);
    termvec[base+2*nports+i] = new int[4];
    termvec[base+2*nports+i][0] = -1;
    termvec[base+2*nports+i][1] = nports+1+i;
    termvec[base+2*nports+i][2] = i;
    termvec[base+2*nports+i][3] = nports; 
    insert(gyr2, termvec[base+2*nports+i]);
    // Time-delayed current source
    ThermalCS* tdcs = new ThermalCS(nports, nsamples-1, time);
    termvec[base+i] = new int[2+2*nports];
    termvec[base+i][0] = nports+1+i;
    termvec[base+i][1] = -1;
    // The rest is set in the following loop
    for (int j=0; j<nports; j++) {
      // Index to locate elements of the thermal step-response matrix
      int index;
      if (j<i)
	index = i*(i+1)/2 + j;
      else
	index = j*(j+1)/2 + i;
      // Continue setting up tdcs
      termvec[base+i][2+2*j] = nports+1+j;
      termvec[base+i][2+2*j+1] = -1;
      // Set the vector to perform the summation
      tdcs->setDeltaRthVec(j, deltaRth + index*(nsamples-1));
      // Insert frequency-domain transconductance
      LinearfVCCS* lvccs = new ThermalVCCS(nfreq, frequency, 
					   Rth + index*nfreq);
      termvec[i*nports+j] = new int[4];
      termvec[i*nports+j][0] = nports+1+i;
      termvec[i*nports+j][1] = -1;
      termvec[i*nports+j][2] = nports+1+j;
      termvec[i*nports+j][3] = -1;
      insert(lvccs, termvec[i*nports+j]); 
    }
    // Now tdcs is ready
    insert(tdcs, termvec[base+i]);
  }
  // The initialization is complete
  flag = true;

}
