#include "global.h"
#include "Vpulse.h"
#include <carrot/Options.h>
#include <carrot/FADBADIndependentCS.h> 
#include <carrot/FADBADLinearVCS.h> 
// Debug
//#include <cstdio>

static ItemInfo Vpulse_info = {
  "vpulse",
  "Pulsed Voltage Source",
  "Carlos Christoffersen",
  DEFAULT_ADDRESS
};

// ***********************************************************************
// Double parameters. We need a way to correlate this with the
// information in SpiceDiode_dparams
// ***********************************************************************
enum {V1, V2, TR, TF, PULSEDT, RI, MAX_DPARM_INDEX}; 

static ParamHandler<double>::ParmInfo Vpulse_dparams[] = {
  {"v1", "Initial value (V)", true, 0.},
  {"v2", "Pulsed value (V)", true, 0.},
  {"tr", "Rise time (s)", true, 0.},
  {"tf", "Fall time (s)", true, 0.},
  {"tp", "Pulsed time (s)", true, 0.},
  {"ri", "Internal resistance (Ohms)", false, 0.}
};

// ***********************************************************************
// Integer parameters
// ***********************************************************************
// enum {TPORT, MAX_IPARM_INDEX};
// static ParamHandler<int>::ParmInfo SpiceDiode_iparams[] = {
//   {"tport", "Use thermal port (4 terminals)", false, 0}
// };


extern "C" {
  //Export the name of the module we are using
  const char * ModuleName(){
    return Vpulse_info.type.c_str();
  }
  ParmSet * newParmSet(const char * model){
    return new ParmSet(model, &Vpulse_info, 
		       Vpulse_dparams, MAX_DPARM_INDEX);
  }
  void delParmSet(ParmSet * p){
    delete p;
    p = 0;
  }
  MacroElement * newObject(ParmSet * p, const char * name){
    return new Vpulse(p, name);
  }
  void delObject(MacroElement * e){
    delete e;
    e = 0;
  }
};

// ***********************************************************************
// Internal CS element definitions:
// ***********************************************************************
// This has to be included after enum (for double params) definition

#include "InternalPulseCS.h"
#include "RIType.h"

// Instantiate template
//template InternalVpulse;

// ***********************************************************************
// SpiceDiode members start:
// ***********************************************************************
Vpulse::Vpulse(ParmSet *pset, const std::string& name) 
  : MacroElement(pset, name, 2)
{
}

Vpulse::~Vpulse()
{
  // There is not much to delete here because MacroElement takes care
}

void Vpulse::init()
{
  double *dparm = getParmSet()->getDHandler()->getParmVec();
  const double &scale = options.getDHandler()->getParmVec()[OPT_SCALE];
  // This flow should be similar for many MacroElements:
  //
  // 0. Check parameters and terminals
  // (We should really throw and exception in the following:)
  getParmSet()->checkParms();
  assert(checkTerms());
  // 1. Create individual CSs
  InternalPulseCS* dccs = new InternalPulseCS(dparm, NULL);
  if (dparm[RI] == 0.) {
    // Need gyrator: topology 1
    // "ri" parameter not used in this case. Perhaps a warning would
    // be appropriate here if ri was set.
    assert(scale != 0.);
    // Create internal terminal (voltage of this is current of voltage source)
    // We also need a ground terminal (not counted as internal)
    setNInternalNodes(1);
    static const int tvec_source[2] = {-1,2};
    insert(dccs, tvec_source);

    // Gyrator sources
    LinearVCCS* g1 = new LinearVCCS(scale);
    static const int tvec_g1[4] = {2,-1,0,1};
    insert(g1, tvec_g1);
    LinearVCCS* g2 = new LinearVCCS(scale);
    static const int tvec_g2[4] = {0,1,2,-1};
    insert(g2, tvec_g2);
  } else {
    // No internal nodes necessary
    setNInternalNodes(0);
    static const int tvec_source[2] = {1,0};
    insert(dccs, tvec_source);

    // No gyrator is necessary: topology 2
    LinearVCCS* gi = new RIType(dparm, NULL);
    static const int tvec_gi[4] = {0,1,0,1};
    insert(gi, tvec_gi);
  }
}

