// This may look like C code, but it is really -*- C++ -*-
//
// Each circuit in the program is contained by an object of type
// Circuit. Subcircuits are managed by a derived class (?).
//
// by Carlos E. Christoffersen

#ifndef Circuit_h
#define Circuit_h 1

#include <cassert>
#include <carrot/MacroElement.h>
#include <carrot/NodeVector.h>
#include <carrot/ObjectLoader.h>
#include <string>

#include <map>

typedef std::map<std::string, ParmSet*> ParmSetMap;
typedef std::map<std::string, MacroElement*> MElemMap;
typedef std::map<std::string, Terminal*> TermMap;

class Circuit 
{
public:

  Circuit(bool libflag = false);
  
  virtual ~Circuit();

  // Get the ModuleMap used by the circuit (used by the parser). This
  // could be hidden from the user but we can not do that with
  // Garrett's parser.
  inline ModuleMap& getModuleMap()
  {
    return elementLibraryMap;
  }

  //--------------------------------------
  // Parameter sets (`models')
  //--------------------------------------

  void addParmSet(ParmSet* pset);

  // Get a parameter set by name (useful for the parser)
  ParmSet* getParmSet(const std::string& name);

  inline ParmSetMap& getParmSetMap()
  {
    return psetMap;
  }

  // Get the maximum number of double parameters of any parmset in
  // circuit
  inline int getMaxDParms() 
  {
    //    assert(uptodate);
    return max_dparms;
  }

  //--------------------------------------
  // Elements 
  //--------------------------------------

  // The following three add one of the fundamental element types to
  // the circuit. They set the circuit vector index in element. The
  // index is (-1) if the element is not already in a circuit.  The
  // remove() functions DO NOT free the CS pointer (except for
  // MacroElements) and no check is made if the CS are still connected
  // to terminals.
  void addElement(LinearVCCS* vccs);
  void remove(LinearVCCS* vccs);

  void addElement(LinearfVCCS* vccs);
  void remove(LinearfVCCS* vccs);

  void addElement(LinearVCQS* vcqs);
  void remove(LinearVCQS* vcqs);

  void addElement(IndependentCS* ics);
  void remove(IndependentCS* ics);

  void addElement(GenericVCCS* vccs);
  void remove(GenericVCCS* vccs);

  void addElement(TimeDelayedCS* vccs);
  void remove(TimeDelayedCS* vccs);

  void addElement(GenericVCQS* vcqs);
  void remove(GenericVCQS* vcqs);

  // Add MacroElements 
  void addElement(MacroElement* me);
  // IMPORTANT: "me" is freed by this function. Use
  // removeFloatingTerms() to clean up any floating terminals left
  // after removal of me.
  void remove(MacroElement* me);
  // Get MacroElement by name
  MacroElement* getMElement(const std::string& name);

  // Functions to get the element vectors
  inline LinVCCSVector& getLinVCCSvec()
  {
    return lCVec.getVec();
  }
  inline LinfVCCSVector& getLinfVCCSvec()
  {
    return flCVec.getVec();
  }
  inline LinVCQSVector& getLinVCQSvec()
  {
    return lQVec.getVec();
  }
  inline IndepCSVector& getIndepCSvec()
  {
    return iCVec.getVec();
  }
  inline GenVCCSVector& getGenVCCSvec()
  {
    return gCVec.getVec();
  }
  inline TDCSVector& getTDCSvec()
  {
    return tdCVec.getVec();
  }
  inline GenVCQSVector& getGenVCQSvec()
  {
    return gQVec.getVec();
  }
  inline MacroElVector& getMElementvec()
  {
    return mVec.getVec();
  }

  // Get the maximum number of controlling voltages for a GenericVVCS
  // The return value is valid only after running setMatrixRCs()
  inline int getMaxCV() const
  {
    assert(uptodate);
    return max_cv;
  }
  // Get the maximum number of current sources in a MacroElement
  // The return value is valid only after running setMatrixRCs()
  inline int getMaxCS() 
  {
    assert(uptodate);
    return max_cs;
  }

  //--------------------------------------
  // Terminals 
  //--------------------------------------

  // Insert a new Terminal
  void addTerminal(Terminal* term);
  // Warning: no check is made if the terminal is still connected to
  // other elements in the circuit. Use with care.
  void remove(Terminal* term);
  // Get Terminal by name
  Terminal* getTerminal(const std::string& name);
  // Remove floating nodes (possibly after a removal of a MacroElement)
  void removeFloatingTerms();

  inline Terminal* getGroundTerm()
  {
    return tVec.getElem(0);
  }

  inline TerminalVector& getTermVec()
  {
    return tVec.getVec();
  }

  inline unsigned getNumberOfTerms() 
  {
    return tVec.size();
  }

  //--------------------------------------
  // General 
  //--------------------------------------

  // Set the row/column field in terminals depending on the elements
  // of the circuit that are considered. (terminals not connected to
  // considered elements are not set). Also set-up quick access
  // function in considered elements.  
  // For now consider always everything
  int setMatrixRCs();


  // Initialise all the elements in circuit
  // This should throw an exception
  void init();
  // Initialise and expand a single MacroElement (for example after
  // the rest of the cricuit was already initialised and expanded).
  void initMacroElement(MacroElement* me);

  // Print the circuit to standard output.
  // Useful for debugging purposes
  void print();

//   //--------------------------------------
//   // Methods to treat subcircuit instances
//   //--------------------------------------

//   // Add one subcircuit to the instance list.
//   // It would be nice if this could be done automagically,
//   // but that would complicate element handling and it is
//   // not such a big deal, I think.
//   void addXinstance(Element* sub_instance)
//     {
//       subckt_list.push_back(sub_instance);
//     }

//   // Expand all subcircuits
//   void expand();  


private:

  // Remove all CS in vector.
  template <typename VEC>
  void eraseCSVector(VEC& vec);

  // Connect all terminals from each CS in vector.
  template <typename VEC>
  void processCSVector(VEC& vec, 
		       TerminalVector& metv, const int& totTerms);

  // Search for terminals connected to one set of elements
  template <typename T>
  void setTermTable(int *termTable, int& nterms, T& vec);
  
  // Set RC info in one set of elements
  template <typename T>
  void setRCInfo(T& vec);

  // Disconnect all terminals from a device
  void disconnect(ElementNode* en);

  // Set RC info in one set of elements. This is an specialisation for
  // GenericVCCS. The difference is that we search for max_cv too.
  void setRCInfo(GenVCCSVector& vec);
  void setRCInfo(TDCSVector& vec);

//   // Check the connectivity of the circuit
//   void checkTerms() throw(char*);

  // Maps to quickly access by name
  ParmSetMap psetMap;
  MElemMap melemMap;
  TermMap termMap;

  // Vectors of elements
  // Currents
  NodeVector<LinearVCCS*, LinVCCSVector> lCVec;
  NodeVector<LinearfVCCS*, LinfVCCSVector> flCVec;
  NodeVector<IndependentCS*, IndepCSVector> iCVec;
  NodeVector<GenericVCCS*, GenVCCSVector> gCVec;
  NodeVector<TimeDelayedCS*, TDCSVector> tdCVec;
  // Charges
  NodeVector<LinearVCQS*, LinVCQSVector> lQVec;
  NodeVector<GenericVCQS*, GenVCQSVector> gQVec;
  // MacroElements
  NodeVector<MacroElement*, MacroElVector> mVec;
  // Vector of terminals
  NodeVector<Terminal*, TerminalVector> tVec;

  // Flag to mark whether the circuit was initialised or not.
  bool init_flag;
  // Maximum number of parameters in a ParmSet
  int max_dparms;
  // Maximum number of controlling voltages for a GenericVCCS
  int max_cv;
  // Maximum number of CS in a MacroElement
  int max_cs;

  // Flag to mark whether max_cv, max_cs and rows and columns are valid.
  bool uptodate;

  // elementLibraryMap is a map to retrieve (or erase) macroelements
  // and parameter sets from the shared libraries.
  ModuleMap elementLibraryMap;

//   //--------------------------------------
//   // Subcircuit-instance related variables
//   //--------------------------------------

//   // vector to hold subcircuit instances
//   ElementVector subckt_list;

//   // Flag to mark when the subcircuits are expanded.
//   bool flattened;


};


#endif

