// This may look like C code, but it is really -*- C++ -*-
//
// Implements a power source using a GenericVCCS (with automatic
// differentiation)
//
// 

#ifndef FADBADPowerSource_H
#define FADBADPowerSource_H

#include <carrot/GenericVCCS.h>
// #include <fadbad/badiff.h>

#include <fadbad/fadiff.h>

// EV: evaluation class
// ncvolt: number of controlling voltages
// ndparms: number of parameters (type double)

template <typename EV, unsigned ncvolt, 
	  unsigned ndparms, int temp_idx>
class FADBADPowerSource : public GenericVCCS {
public:

  // dparm and iparm values are maintained by the MacroElement
  // (here we assume these are ready to use)
  FADBADPowerSource(double* dparm, int* iparm)
    : GenericVCCS(ncvolt), dparm(dparm), teval(iparm) {}

  virtual ~FADBADPowerSource() {}

  // Call this each time the parameter values are changed
  // (no need for virtual here?)
  virtual void init()
  {
    // Check that terminals are connected.  Only needed for
    // debugging. Properly written MacroElements should take care of
    // this.
    assert(ElementNode::checkTerms());
    // Adjust teval parameters to temperature in parameter list
    teval.newParms(cvar, dparm);
    teval.setTemp(tvar, dparm, cvar, dparm[temp_idx]);
  }

  // These are defined in the base classes
  virtual void eval(double& f, double* v)
  {
    // Set temperature first
    teval.setTemp(tvar, dparm, cvar, v[ncvolt-1]);
    teval(f, v, dparm, tvar, cvar);
    teval.power(f, v, dparm);
  }

  virtual void eval_and_deriv(double& f, double* df, double* v);
  virtual bool paramDeriv(double* df, double *v);

private:

  // Parameter definitions
  double *dparm;
  double tvar[EV::MAX_TVAR_INDEX];
  double cvar[EV::MAX_CVAR_INDEX];
  EV teval;
};


template <typename EV, unsigned ncvolt, 
	  unsigned ndparms, int temp_idx>
void FADBADPowerSource<EV, ncvolt, ndparms, temp_idx>::
eval_and_deriv(double& f, double* df, double* v)
{
  // The last controlling voltage is the temperature
  unsigned ncvolt1(ncvolt - 1);
  F<double> fv[ncvolt], ftvar[EV::MAX_TVAR_INDEX], ff;
  for (unsigned i(0); i < ncvolt; i++) {
    fv[i] = v[i];
    fv[i].diff(i, ncvolt); // differenciate respect to v[i]
  }
  // Consider temperature dependencies (fv[ncvolt1] = temperature)
  teval.setTemp(ftvar, dparm, cvar, fv[ncvolt1]);
  // evaluate function
  teval(ff, fv, dparm, ftvar, cvar);   
  teval.power(ff, fv, dparm);
  f = ff.x(); // Function value
  for (unsigned i(0); i < ncvolt; i++)
    df[i] = ff.d(i); // df/dv[i]
}

template <typename EV, unsigned ncvolt, 
	  unsigned ndparms, int temp_idx>
bool FADBADPowerSource<EV, ncvolt, ndparms, temp_idx>::
paramDeriv(double* df, double *v)
{
  // Derivative with respect to temperature parameter is zero
  // (because the independent variable is now the last control
  // voltage).
  F<double> fdparm[ndparms], 
    ftvar[EV::MAX_TVAR_INDEX], fcvar[EV::MAX_CVAR_INDEX], ff;
  for (unsigned i(0); i < ndparms; i++) {
    fdparm[i] = dparm[i];
    fdparm[i].diff(i, ndparms); // differenciate respect to one parameter
  }
  // Consider variables not dependent on temperature or voltage
  teval.newParms(fcvar, fdparm);
  // Consider temperature dependencies (use parameter temperature)
  teval.setTemp(ftvar, fdparm, fcvar, v[ncvolt-1]);
  // evaluate function
  teval(ff, v, fdparm, ftvar, fcvar);   
  teval.power(ff, v, fdparm);
  for (unsigned i(0); i < ndparms; i++) {
    df[i] = ff.d(i);
  }
  return true;
}


  // Experiment with backward differentiation. It did not work due
  // to bugs in badiff.h.

//     B<double> fdparm[ndparms], 
//       ftvar[EV::MAX_TVAR_INDEX], fcvar[EV::MAX_CVAR_INDEX], ff;
//     for (unsigned i(0); i < ndparms; i++) {
//       fdparm[i] = dparm[i];
//     }
//     // Consider variables not dependent on temperature or voltage
//     teval.newParms(fcvar, fdparm);
//     // Consider temperature dependencies (use parameter temperature)
//     teval.setTemp(ftvar, fdparm, fcvar, fdparm[EV::Temp]);
//     // evaluate function
//     teval(ff, v, fdparm, ftvar, fcvar);   
//     ff.diff(0,1);
//     for (unsigned i(0); i < ndparms; i++) {
//       df[i] = (fdparm[i]).d(0);
//     }




#endif
