// This may look like C code, but it is really -*- C++ -*-
//
// GenericVCCS is the class that handles all (nonlinear, not sources)
// voltage-controlled current-sources (the basic block of a circuit)
// in the program:
//
//          2      
//          o                        o   o   o         o      
//          |                                                
//          |      	             +   +   +         +     
//         (^) i(v1,v2,v3,...,vn)    v1  v2  v3        vn
//          |      	             -   -   -         -     
//          |                                              
//          o      	             o   o   o         o     
//          1      
//
// a GenericVCCS have at least 2 terminals (1 and 2).
// 
// by Carlos E. Christoffersen

#ifndef GenericVCCS_H
#define GenericVCCS_H 1

#include <carrot/ElementNode.h>
#include <carrot/Terminal.h>
#include <carrot/InternalCS.h>

class GenericVCCS : public ElementNode, public InternalCS
{
public:

  // ElementNode takes care of the connectivity.  ncvolt is the number
  // of controlling voltages. Number of terminals is: 2 output
  // terminals + 2*number of controlling terminals.  
  GenericVCCS(const int& ncvolt) 
    : ElementNode(2+2*ncvolt), InternalCS(), 
  ncv(ncvolt), row1(-1), row2(-1), colplus(NULL), colminus(NULL) {}
  
  virtual ~GenericVCCS();

  // Each derived class implements f(v1,v2,...) differently:
  virtual void eval(double& f, double* v) = 0;

  // This returns the derivatives of f with respect to all v:
  // df/dv1, df/dv2, df/dv3, ...
  virtual void eval_and_deriv(double& f, double* df, double* v) = 0;

  // Given the controlling voltages, return the derivative of the current 
  // with respect to all double parameters
  virtual bool paramDeriv(double* df, double *v)
  {
    // false means not implemented
    return false;
  }

  // Functions for quick access to matrix row and column information
  inline const int& getRow1() const
  {
    return row1;
  }

  inline const int& getRow2() const
  {
    return row2;
  }

  inline const int* getColPlus() const
  {
    return colplus;
  }

  inline const int* getColMinus() const
  {
    return colminus;
  }

  inline const int& getNCVolt() const
  {
    return ncv;
  }

  void setRCinfo();

private:

  int ncv;
  int row1;
  int row2;
  int *colplus;
  int *colminus;

};

typedef std::vector<GenericVCCS*> GenVCCSVector;

#endif
