// This may look like C code, but it is really -*- C++ -*-
//
// LinearfVCCS is the class that handles linear voltage-controlled
// current-sources in frequency domain.
//
//          1                        3
//          o                        o       
//          |                               
//          |                        +      
//         (^) g v1                  v1  
//          |                        -      
//          |                             
//          o                        o      
//          2                        4
//
// If there are more than one controlling voltage we just add 2 or
// more LinearfVCCS in parallel.
//
// by Carlos E. Christoffersen

#ifndef LinearfVCCS_h
#define LinearfVCCS_h 1

#include <carrot/ElementNode.h>
#include <carrot/Terminal.h>
#include <carrot/InternalCS.h>
#include <complex>

typedef std::complex<double> double_complex;

class LinearfVCCS : public ElementNode, public InternalCS
{
public:

  LinearfVCCS() 
    : ElementNode(4), InternalCS(), 
      row1(-1), row2(-1), col1(-1), col2(-1) {}

  virtual ~LinearfVCCS() {}
    
  // This returns the gain of the VCCS. That is all we really need
  // (besides the terminal connections).
  virtual void getGain(double_complex& y, const double& f) = 0;

  // Functions for quick access to matrix row and column information
  inline const int& getRow1() const
  {
    return row1;
  }

  inline const int& getRow2() const
  {
    return row2;
  }

  inline const int& getCol1() const
  {
    return col1;
  }

  inline const int& getCol2() const
  {
    return col2;
  }

  inline void setRCinfo()
  {
    row1 = getTerminal(0)->getRC();
    row2 = getTerminal(1)->getRC();
    col1 = getTerminal(2)->getRC();
    col2 = getTerminal(3)->getRC();
  }

  // Return the derivative of g with respect to all double parameters
  virtual bool paramDeriv(double* dg) 
  {
    // False means not implemented
    return false;
  };

private:
  
  int row1, row2, col1, col2;

};

typedef std::vector<LinearfVCCS*> LinfVCCSVector;

#endif


