// This may look like C code, but it is really -*- C++ -*-
//
// Macroelement is the class for all elements available to the user.
//
// A Macroelement can have any number of external terminals (at least
// 2). There may be internal terminals too.
// 
// 

#ifndef MacroElement_H
#define MacroElement_H

#include <carrot/LinearVCCS.h>
#include <carrot/LinearfVCCS.h>
#include <carrot/GenericVCCS.h>
#include <carrot/TimeDelayedCS.h>
#include <carrot/LinearVCQS.h>
#include <carrot/GenericVCQS.h>
#include <carrot/IndependentCS.h>
#include <carrot/NamedItem.h>
#include <carrot/ParmSet.h>

class Circuit;

class MacroElement : public ElementNode , public NamedItem
{
public:

  // ElementNode takes care of the connectivity 
  // ParmSet takes care of parameter handling
  MacroElement(ParmSet* pset1, const std::string& name, unsigned nterms = 0) 
    : ElementNode(nterms), NamedItem(name), pset(pset1), 
      lvcCSvec(), flvcCSvec(), lvcQSvec(), 
      genVCCSvec(), tdCSvec(), genVCQSvec(), 
      indepCSvec(), nInternalNodes(-1)
  {
    pset->addME(this);
  }  

  virtual ~MacroElement();

  // Deal with internal nodes (terminals)
  inline void setNInternalNodes(const int& n)
  {
    nInternalNodes = n;
  }
  inline const int& getNInternalNodes() const
  {
    assert(nInternalNodes > -1);
    return nInternalNodes;
  }

  // Initialization: used for taping, error checking, etc.
  // Defaults to nothing.
  // This should be called after any changes in parameters.
  virtual void init() {}

  // Initialise all CS in cSvec.  
  // Called right after init() by the Circuit class
  void initAllCS();

  // This is to make sure we can access parameters from any
  // macroelelement.
  inline ParmSet* getParmSet()
  {
    return pset;
  }

  // Insert individual CS in respective vectors
  inline void insert(LinearVCCS* cs, const int* val_vec) 
  {
    cs->setTermNum(val_vec);
    lvcCSvec.push_back(cs);
  }
  inline void insert(LinearfVCCS* cs, const int* val_vec) 
  {
    cs->setTermNum(val_vec);
    flvcCSvec.push_back(cs);
  }
  inline void insert(LinearVCQS* cs, const int* val_vec) 
  {
    cs->setTermNum(val_vec);
    lvcQSvec.push_back(cs);
  }
  inline void insert(GenericVCCS* cs, const int* val_vec) 
  {
    cs->setTermNum(val_vec);
    genVCCSvec.push_back(cs);
  }
  inline void insert(TimeDelayedCS* cs, const int* val_vec) 
  {
    cs->setTermNum(val_vec);
    tdCSvec.push_back(cs);
  }
  inline void insert(GenericVCQS* cs, const int* val_vec) 
  {
    cs->setTermNum(val_vec);
    genVCQSvec.push_back(cs);
  }
  inline void insert(IndependentCS* cs, const int* val_vec) 
  {
    cs->setTermNum(val_vec);
    indepCSvec.push_back(cs);
  }

  // Returns all linear VCCS in this macroelement
  inline LinVCCSVector& getLinearVCCSvec()
  {
    return lvcCSvec;
  }
  // Returns all time-varying linear VCCS in this macroelement
  inline LinfVCCSVector& getLinearfVCCSvec()
  {
    return flvcCSvec;
  }
  // Returns all linear VCQS in this macroelement
  inline LinVCQSVector& getLinearVCQSvec()
  {
    return lvcQSvec;
  }
  // Returns all generic VCCS in this macroelement
  inline GenVCCSVector& getGenVCCSvec()
  {
    return genVCCSvec;
  }
  // Returns all time-delayed VCCS in this macroelement
  inline TDCSVector& getTDCSvec()
  {
    return tdCSvec;
  }
  // Returns all generic VCQS in this macroelement
  inline GenVCQSVector& getGenVCQSvec()
  {
    return genVCQSvec;
  }
  // Returns all independent CS in this macroelement
  inline IndepCSVector& getIndepCSvec()
  {
    return indepCSvec;
  }

  // Returns the total number of CS
  inline int getCSNumber()
  {
    return lvcCSvec.size() + flvcCSvec.size()+ lvcQSvec.size() 
      + genVCCSvec.size() + tdCSvec.size() + genVCQSvec.size() 
      + indepCSvec.size();
  }

private:

  // Remove CS or terminals in one vector from circuit
  template <typename VECTYPE>
  void clearItems(VECTYPE &vec);

  // Parameter set (Spice `.model') pointer
  ParmSet* pset;
  // Vectors with linear CS and QS only. Required to identify
  // derivatives of conductances with respect to parameters.
  LinVCCSVector lvcCSvec;
  LinfVCCSVector flvcCSvec;
  LinVCQSVector lvcQSvec;
  // Vectors for the other types of CS
  GenVCCSVector genVCCSvec;
  TDCSVector tdCSvec;
  GenVCQSVector genVCQSvec;
  IndepCSVector indepCSvec;
  int nInternalNodes;
};

typedef std::vector<MacroElement*> MacroElVector;

#endif
