// This may look like C code, but it is really -*- C++ -*-
//
// This is a template class to perform the Newton iterations. Only the
// iteration control is performed here, the matrix and rhs vector are
// calculated elsewhere.
//

#ifndef NewtonSolve_H
#define NewtonSolve_H

#include <carrot/Options.h>
#include <carrot/UMFPackDI.h>
#include <cmath>
#include <cstdio>

//#define DEBUG 1

template <typename EvalClass>
class NewtonSolve 
{
public:

  // iSize is the initial estimation of the matrix size.
  NewtonSolve(EvalClass* ec, UMFPackDI* nam_p, const int& nrcs)
    : ec(ec), nam_p(nam_p), nrcs(nrcs)  
  { 
    // Source vector (with ground)
    svec_ground = new double[nrcs+1]; 
    // leave extra ground element out (svec[-1])
    svec = svec_ground + 1; 
    // Previous iteration nodal voltages
    uvec_j_ground = new double[nrcs+1];
    uvec_j = uvec_j_ground + 1;
    // Backup to go back Newton iterations
    uvec_backup = new double[nrcs];
  }

  ~NewtonSolve()
  {
    delete [] uvec_backup;
    delete [] uvec_j_ground;
    delete [] svec_ground;
  }

  // Solve the nonlinear system. The initial guess must be passed in
  // uvec. indep_svec is the vector with the contributions from
  // independent sources only (that should be constant during Newton
  // Iterations).
  // iter and rejections are output parameters
  double solve(double *uvec, double *indep_svec,
	       int& iter, int& rejections)
  {
    // Set initial values
    iter = 0;
    rejections = 0;
    // Copy initial guess
    for (int i(0); i<nrcs; i++)
      uvec_j[i] = uvec[i];
    // Set potential of reference node
    uvec_j_ground[0] = 0.;
    double residual(1.);
    double *dparm = options.getDHandler()->getParmVec();
    int *iparm = options.getIHandler()->getParmVec();
    double &tol = dparm[OPT_TOL];
    int local_iter(0);
    int &max_iter = iparm[OPT_MAXITER];
    double ss_factor = dparm[OPT_ISSF];
    bool verbose = iparm[OPT_NEWTVERB];
    double beta = 1.;
    bool reject(false);
    double old_ss = ss_factor;
    double gmin = 0.;
    // Try no source stepping first
    ss_factor = 1.;
    // Save initial guess
    for (int i(0); i<nrcs; i++) 
      uvec_backup[i] = uvec_j[i];
    bool first_try = true;

    // *********************************************************************
    // Main Loop follows
    // *********************************************************************
    while ((residual > tol || ss_factor < 1. || gmin > 0.) 
	   && iter < max_iter) {
      if (first_try) {
	if (verbose)
	  printf("t");
	if ((local_iter > 10 && residual > .5) || reject) {
	  // no luck, guess was not close to solution, restore old
	  ss_factor = old_ss;
	  // If we are going to try source stepping, we should start
	  // from zero here:
	  for (int i(0); i<nrcs; i++) 
	    uvec_j[i] = 0.;
	    //	    uvec_j[i] = uvec_backup[i];
	  first_try = false;
	  reject = false;
	  local_iter = 0;
	  residual = 1.;
	  beta = 4.;
	  gmin = 1e-6;
	}
      }
      if (ss_factor == 1. && residual < 1e-2 && gmin > 0.) {
	// Reduce gmin
	if (gmin > 1e-8)
	  gmin *= .8;
	else if (gmin > 1e-12)
	  gmin *= .1;
	else
	  gmin = 0.;
	if (verbose)
	  printf("g");
	local_iter = 0;
	beta = 1.;
      }
      if (reject || local_iter > 30) {
	rejections++;
	// We have to go back
	ss_factor = .1 * (9. * old_ss + ss_factor);
	// Restore old good values
	for (int i(0); i<nrcs; i++) 
	  uvec_j[i] = uvec_backup[i];

	reject = false;
	residual = 1.;
	local_iter = 0;
	if (verbose)
	  printf("x");
      }
      if (ss_factor != 1.) {
	if (residual < 1e-3 && beta < 2.) {
	  old_ss = ss_factor;
	  // Test: do not use fixed source stepping, adjust the size
	  // according to source sensitivity.
	  nam_p->solve(uvec, indep_svec);
	  // Calculate infinity norm
	  double norm = 0.;
	  for (int i(0); i<nrcs; i++) {
	    double value = fabs(uvec[i]);
	    if (value > norm)
	      norm = value;
	  }
	  // max_delta is the maximum increment that we allow in the nodal
	  // vector.  We need to adapt this value if progress is good. For
	  // example count the number of iterations required for last
	  // step. If the step fails, go back and reduce max_delta.
	  double max_delta;
	  if (local_iter < 5) {
	    max_delta = .4; 
	    beta = 2.;
	  }
	  else {
	    max_delta = .02;
	    beta = 10.;
	  }
	  ss_factor += max_delta / norm;
	  //       ss_factor *= dparm[SS_INC];
	  if (ss_factor/old_ss > 3.)
	    ss_factor = old_ss * 3.;
	  if (ss_factor > 1.) {
	    ss_factor = 1.;
	    // Make sure we use beta=1 in the last iteration
	    beta = 1.;
	  }
	  // Now update uvec with a better guess (and save old values)
	  for (int i(0); i<nrcs; i++) {
	    uvec_backup[i] = uvec_j[i];
	    uvec_j[i] += uvec[i] * (ss_factor - old_ss);
	  }
	  // An alternative to check if this step value is fine would be to
	  // find if the derivative values at the new guess are too
	  // different from the derivatives at the previous source
	  // step. If they are, reduce the step size until the difference
	  // is not too large.

	  // Reset local iterations
	  local_iter = 0;
	  if (verbose)
	    printf("s");
	} else {
	  if (residual > 10.) {
	    // Restore old good values
	    for (int i(0); i<nrcs; i++) 
	      uvec_j[i] = uvec_backup[i];
	    // Increase beta
	    beta *= residual;
	    if (beta > 1e2)
	      beta = 1e2;
	    rejections++;
	  }
	  else {
	    beta *= .5;
	    if (beta < 1.)
	      beta = 1.;
	  }
	}
      }

      if (verbose)
	printf(".");
      // Clear numeric structure and nonlinear part from NAM
      nam_p->clearNonlinearNumeric();

      // Restore source vector with independent source contributions
      for (int i(0); i<nrcs; i++){
	svec[i] = ss_factor * indep_svec[i];
      }
      
      // retcode values:
      // 0: normal processing
      // 1: Problem is linear, no iterations necessary
      // >=2: overflow or some other nasty condition
      int retcode = ec->fillNAMandRHS(*nam_p, svec, uvec_j, beta, gmin);
      // If something went wrong just discard everything
      if (retcode > 1) {
	reject = true;
	residual = 1.;
	iter++;
	local_iter++;
	continue;
      }
    
#ifdef DEBUG
      nam_p->printTriplet();
      //exit(0);
#endif
      // Factor matrix and solve system.
      nam_p->buildCompressed();
      // The symbolic object is always the same, create it once
      // (the nam takes care of this internally)
      nam_p->symbolic();

      if (nam_p->factor() || nam_p->solve(uvec, svec)) {
	reject = true;
	residual = 1.;
	iter++;
	local_iter++;
	continue;
      }

      // residual == 0. if retcode == 1 (linear problem)
      residual = 0.;
      if (retcode == 0) {
	// Update uvec_j here and calculate residual. Use infinity norm.
	for (int i(0); i<nrcs; i++) {
	  double diff = fabs(uvec[i] - uvec_j[i]);
	  if (diff > residual)
	    residual = diff;
	  // Save results
	  uvec_j[i] = uvec[i];
#ifdef DEBUG
	  printf("v[%d] = %g\n", i, uvec[i]);
#endif
	}
      }

      iter++;
      local_iter++;
    }
    // Beta must be one for optimum convergence and also we need this for
    // sensitivity calculation.
    if (iter < max_iter) {
      assert(beta == 1.);
      assert(gmin == 0.);
    }
    return residual;
  }

private:

  EvalClass* ec;
  UMFPackDI* nam_p;
  int nrcs;
  // RHS vector
  double *svec_ground;
  double *svec;
  // Previous iteration nodal voltages
  double *uvec_j_ground;
  double *uvec_j;
  // Backup to go back Newton iterations
  double *uvec_backup;

};

#endif
